from __future__ import annotations

import datetime
import importlib
import string
import unicodedata

import iotbx.cif.model


def get_beamline_definition(detector_id, **kwargs):
    if isinstance(detector_id, bytes):
        detector_id = detector_id.decode("utf-8", "ignore")

    valid_chars = frozenset(f"_.{string.ascii_letters}{string.digits}")
    filename = unicodedata.normalize("NFKD", detector_id)
    filename = "".join(c if c in valid_chars else "_" for c in filename)
    while "__" in filename:
        filename = filename.replace("__", "_")
        # http://stackoverflow.com/questions/1546226/a-simple-way-to-remove-multiple-spaces-in-a-string-in-python/15913564#15913564

    try:
        beamline = importlib.import_module("dxtbx.data.beamline_defs.%s" % filename)
        generator_object = beamline.get_definition(**kwargs)
    except ImportError:
        generator_object = Dummy()
    generator_object.set_detector_name(detector_id)
    generator_object.set_block_name(filename)
    return generator_object


class template:
    def CIF_block(self):
        """Interface function to generate a CIF block for this detector."""
        raise RuntimeError("This needs to be overridden.")

    def mmCIF_block(self):
        """Interface function to generate an mmCIF block for this detector."""
        raise RuntimeError("This needs to be overridden.")

    def _lookup(self, mmCIFsemantics):
        keys = {
            # Values can either be
            #
            # a tuple
            #    ( CIF-string, mmCIF-string )
            #
            # or a string such as
            #    "_something?something_else"
            # where
            #   _something_something_else is CIF, and
            #   _something.something_else is mmCIF
            #
            # Do NOT use underscores in keys.
            "df.detector": "_diffrn?detector",
            "df.detector.type": "_diffrn_detector?type",
            "df.m.dev": "_diffrn?measurement_device",
            "df.m.dev.type": "_diffrn?measurement_device_type",
            "df.m.method": "_diffrn?measurement_method",
            "df.m.spec.supp": "_diffrn?measurement_specimen_support",
            "df.rad.mono": "_diffrn_radiation?monochromator",
            "df.src": ("_diffrn_source", "_diffrn_source.source"),
            "df.src.details": "_diffrn_source?details",
            "df.src.type": "_diffrn_source?type",
            "references": "_publ?section_references",
            "sw.collection": "_computing?data_collection",
        }
        column = 1 if mmCIFsemantics else 0

        def __lookup(key):
            entry = keys[key]
            if isinstance(entry, str):
                return entry.replace("?", "." if mmCIFsemantics else "_")
            else:
                return entry[column]

        return __lookup

    def get_block_name(self):
        return self._block_name

    def set_block_name(self, block_name):
        self._block_name = block_name

    def get_detector_name(self):
        return self._detector_name

    def set_detector_name(self, detector_name):
        self._detector_name = detector_name

    def write_block_to_file(self, block, filename):
        c = iotbx.cif.model.cif()
        c[self.get_block_name()] = block
        with open(filename, "w") as fh:
            c.show(out=fh)

    def _date_to_epoch(self, year, month, day):
        return (
            datetime.datetime(year, month, day, 0, 0) - datetime.datetime(1970, 1, 1)
        ).total_seconds()

    def _generate_block(self):
        return iotbx.cif.model.block()

    def __str__(self):
        return "CIF block generator for %s (%s.py)" % (
            self._detector_name,
            self._block_name,
        )


class Dummy(template):
    def __init__(self):
        self.CIF_block = self._generate_block
        self.mmCIF_block = self._generate_block

    def __str__(self):
        return "Dummy CIF generator. No information for detector %s (%s.py)" % (
            self._detector_name,
            self._block_name,
        )



from __future__ import annotations

import shutil

import pytest

from dxtbx.sequence_filenames import template_image_range, template_regex


@pytest.mark.parametrize(
    "filename,template,digits",
    [
        ("foo_bar_001.img", "foo_bar_###.img", 1),
        ("foo_bar001.img", "foo_bar###.img", 1),
        ("foo_bar_1.8A_004.img", "foo_bar_1.8A_###.img", 4),
        ("foo_bar.001", "foo_bar.###", 1),
        ("foo_bar_002.img1000", "foo_bar_###.img1000", 2),
        ("foo_bar_00005.img", "foo_bar_#####.img", 5),
        ("image0010", "image####", 10),
        ("foo_123_1_1.rodhypix", "foo_123_1_#.rodhypix", 1),  # Rigaku-style
    ],
)
def test_template_regex(filename, template, digits):
    assert template_regex(filename) == (template, digits)


def test_template_image_range(dials_data):
    template = str(dials_data("insulin") / "insulin_1_###.img")
    assert template_image_range(template) == (1, 45)


def test_template_image_range_non_zero_padded(dials_data, tmp_path):
    images = sorted(dials_data("insulin").glob("insulin_1_0[0-1]*"))
    # symlink if possible, copy if necessary
    for i, image in enumerate(images):
        try:
            (tmp_path / f"insulin_1_{i + 1}.img").symlink_to(image)
        except OSError:
            shutil.copy(image, (tmp_path / f"insulin_1_{i + 1}.img"))

    template = str(tmp_path / "insulin_1_#.img")
    assert template_image_range(template) == (1, 19)



from __future__ import annotations

import pytest

from dxtbx.format.FormatHDF5EigerNearlyNexusSPring8 import (
    FormatHDF5EigerNearlyNexusSPring8,
)
from dxtbx.model.experiment_list import ExperimentListFactory
from dxtbx.model.goniometer import Goniometer


def test_spring8_ccp4_2018_zenodo_1443110_data03(dials_data):
    master_h5 = (
        dials_data("spring8_ccp4_2018")
        / "ccp4school2018_bl41xu"
        / "05"
        / "data03"
        / "data03_master.h5"
    )
    assert FormatHDF5EigerNearlyNexusSPring8.understand(master_h5)

    expts = ExperimentListFactory.from_filenames([master_h5])
    imageset = expts[0].imageset
    assert imageset.get_format_class() == FormatHDF5EigerNearlyNexusSPring8

    detector = imageset.get_detector()
    gonio = imageset.get_goniometer()
    scan = imageset.get_scan()
    beam = imageset.get_beam()

    panel = detector[0]
    assert panel.get_pixel_size() == pytest.approx((0.075, 0.075))
    assert panel.get_image_size() == (4150, 4371)
    assert panel.get_trusted_range() == (0, 2.094707e06)
    assert panel.get_fast_axis() == (1, 0, 0)
    assert panel.get_slow_axis() == (0, -1, 0)
    assert panel.get_origin() == pytest.approx((-151.939, 169.629, -180), abs=1e-3)
    assert panel.get_distance() == pytest.approx(180)

    assert isinstance(gonio, Goniometer)
    assert gonio.get_rotation_axis() == (-1, 0, 0)
    assert gonio.get_fixed_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)
    assert gonio.get_setting_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)

    assert scan.get_oscillation() == pytest.approx((-10, 1))
    assert scan.get_image_range() == (1, 180)

    assert beam.get_wavelength() == pytest.approx(1.28241, abs=1e-5)
    assert beam.get_s0() == pytest.approx((0, 0, -1 / beam.get_wavelength()))



from __future__ import annotations

import collections
import itertools
import math
import os
from collections.abc import Iterable
from typing import Union

import h5py
import hdf5plugin  # noqa: F401
import numpy

import cctbx.uctbx
from cctbx.eltbx import attenuation_coefficient
from scitbx import matrix
from scitbx.array_family import flex
from scitbx.matrix import col, sqr

import dxtbx.model
from dxtbx.model import (
    Beam,
    Crystal,
    Detector,
    Panel,
    ParallaxCorrectedPxMmStrategy,
    Scan,
    Spectrum,
)

try:
    try:
        from ..dxtbx_format_nexus_ext import (
            dataset_as_flex_double,
            dataset_as_flex_float,
            dataset_as_flex_int,
        )
    except ModuleNotFoundError:
        from dxtbx_format_nexus_ext import (  # type: ignore
            dataset_as_flex_double,
            dataset_as_flex_float,
            dataset_as_flex_int,
        )
except ImportError:
    # Workaround for psana build, which doesn't link HDF5 properly
    if "SIT_ROOT" not in os.environ:
        raise

NXNode = Union[h5py.File, h5py.Group]


def h5str(h5_value: str | numpy.bytes_ | bytes) -> str:
    """
    Convert a value returned an h5py attribute to str.

    h5py can return either a bytes-like (numpy.bytes_) or str object
    for attribute values depending on whether the value was written as
    fixed or variable length. This function collapses the two to str.
    """
    if isinstance(h5_value, (numpy.bytes_, bytes)):
        return h5_value.decode("utf-8")
    return h5_value


def dataset_as_flex(dataset, selection):
    if numpy.issubdtype(dataset.dtype, numpy.integer):
        return dataset_as_flex_int(dataset.id.id, selection)
    else:
        assert numpy.issubdtype(dataset.dtype, numpy.floating)
        double_types = [
            numpy.double,
            numpy.longdouble,
            numpy.float64,
        ]
        if hasattr(numpy, "float96"):
            double_types.append(numpy.float96)
        if hasattr(numpy, "float128"):
            double_types.append(numpy.float128)
        if dataset.dtype in [
            numpy.half,
            numpy.single,
            numpy.float64,
            numpy.float16,
            numpy.float32,
        ]:
            return dataset_as_flex_float(dataset.id.id, selection)
        elif dataset.dtype in double_types:
            return dataset_as_flex_double(dataset.id.id, selection)
        else:
            assert False, "unknown floating data type (%s)" % str(dataset.dtype)


class NXValidationError(RuntimeError):
    """A specific exception to record validation errors"""


def find_classes(node: NXNode, *nx_classes: str | None) -> tuple[list[h5py.Group], ...]:
    """
    Find instances of multiple NXclass types within the children of the current node.

    Args:
        node: The input h5py node (h5py.File or h5py.Group).
        nx_classes: Names of NXclass types to search for.  If None, search for children
            without an NXclass.

    Returns:
        A list of matching nodes for each of the specified NX_class types.
    """
    results: dict[str | None, list[h5py.Group]] = {
        nx_class: [] for nx_class in nx_classes
    }

    values: Iterable[h5py.Group] = filter(None, node.values())
    for v in values:
        class_name = h5str(v.attrs.get("NX_class"))
        if class_name in nx_classes:
            results[class_name].append(v)

    return tuple(results.values())


def find_class(node: NXNode, nx_class: str | None) -> list[h5py.Group]:
    """
    Find instances of a single NXclass type within the children of the current node.

    This is a convenience function, equivalent to calling find_classes with a single
    NXclass type name argument and returning the list of matches.

    Args:
        node: The input h5py node (h5py.File or h5py.Group).
        nx_class: Names of NXclass type to search for.  If None, search for children
            without an NXclass.

    Returns:
        The list of matching nodes for the specified NXclass type.
    """
    return find_classes(node, nx_class)[0]


def find_entries(nx_file: h5py.File) -> list[h5py.Group]:
    """
    Find NXmx entries.

    A valid NeXus file should contain one or more NXentry groups at the top level. Only
    return those NXentry groups that are specified as following the NXmx definition.

    https://manual.nexusformat.org/classes/base_classes/NXentry.html#nxentry

    Args:
        nx_file (h5py.File): The input h5py.File instance.

    Returns:
        list: The list of found NXentry groups.

    """
    return [
        group
        for group in find_class(nx_file, "NXentry")
        if "definition" in group and h5str(group["definition"][()]) == "NXmx"
    ]


def convert_units(value, input_units, output_units):
    """
    Hacky utility function to convert units
    """
    if isinstance(input_units, bytes):
        input_units = input_units.decode("latin-1")
    converters = {
        "m": {
            "mm": lambda x: x * 1e3,
            "microns": lambda x: x * 1e6,
            "nm": lambda x: x * 1e9,
        },
        "mm": {
            "m": lambda x: x * 1e-3,
            "microns": lambda x: x * 1e3,
            "nm": lambda x: x * 1e6,
        },
        "microns": {
            "m": lambda x: x * 1e-6,
            "mm": lambda x: x * 1e-3,
            "nm": lambda x: x * 1e3,
        },
        "nm": {
            "m": lambda x: x * 1e-9,
            "mm": lambda x: x * 1e-6,
            "microns": lambda x: x * 1e-3,
            "angstroms": lambda x: x * 10,
        },
        "angstroms": {"angstrom": lambda x: x},
    }
    if input_units == output_units:
        return value
    try:
        return converters[input_units][output_units](value)
    except Exception:
        raise RuntimeError(f"Can't convert units {input_units!r} to {output_units!r}")


def visit_dependencies(nx_file, item, visitor=None):
    """
    Walk the dependency chain and call a visitor function
    """
    dependency_chain = set()
    if os.path.basename(item) == "depends_on":
        depends_on = nx_file[item][()]
    else:
        depends_on = h5str(nx_file[item].attrs["depends_on"])

    while not depends_on == ".":
        if visitor:
            visitor(nx_file, depends_on)
        if depends_on in dependency_chain:
            raise RuntimeError("'%s' is a circular dependency" % depends_on)
        try:
            _ = nx_file[depends_on]
        except Exception:
            raise RuntimeError("'%s' is missing from nx_file" % depends_on)
        dependency_chain.add(depends_on)
        try:
            depends_on = h5str(nx_file[depends_on].attrs["depends_on"])
        except Exception:
            raise RuntimeError("'%s' contains no depends_on attribute" % depends_on)


def construct_vector(nx_file, item, vector=None):
    """
    Walk the dependency chain and create the absolute vector
    """

    class TransformVisitor:
        def __init__(self, vector):
            self.vector = matrix.col(vector)

        def visit(self, nx_file, depends_on):
            item = nx_file[depends_on]
            value = item[()]
            units = h5str(item.attrs["units"])
            ttype = h5str(item.attrs["transformation_type"])
            vector = matrix.col(item.attrs["vector"])
            if ttype == "translation":
                value = convert_units(value, units, "mm")
                if hasattr(value, "__iter__") and len(value) == 1:
                    value = value[0]
                self.vector = vector * value + self.vector
            elif ttype == "rotation":
                if hasattr(value, "__iter__") and len(value):
                    value = value[0]
                if units == "rad":
                    deg = False
                elif units == "deg":
                    deg = True
                else:
                    raise RuntimeError("Invalid units: %s" % units)
                self.vector.rotate_around_origin(axis=vector, angle=value, deg=deg)
            else:
                raise RuntimeError("Unknown transformation_type: %s" % ttype)

    if vector is None:
        value = nx_file[item][()]
        units = h5str(nx_file[item].attrs["units"])
        ttype = h5str(nx_file[item].attrs["transformation_type"])
        vector = nx_file[item].attrs["vector"]
        if "offset" in nx_file[item].attrs:
            offset = nx_file[item].attrs["offset"]
            offset = convert_units(offset, units, "mm")
        else:
            offset = vector * 0.0
        if ttype == "translation":
            value = convert_units(value, units, "mm")
            try:
                vector = vector * value
            except ValueError:
                vector = vector * value[0]
            vector += offset

    visitor = TransformVisitor(vector)
    visit_dependencies(nx_file, item, visitor.visit)
    return visitor.vector


def construct_axes(nx_file, item, vector=None):
    """
    Walk the dependency chain and create the absolute vector
    """

    class Visitor:
        def __init__(self):
            self._axes = flex.vec3_double()
            self._angles = flex.double()
            self._axis_names = flex.std_string()
            self._is_scan_axis = flex.bool()

        def visit(self, nx_file, depends_on):
            item = nx_file[depends_on]
            value = item[()]
            units = h5str(item.attrs["units"])
            ttype = h5str(item.attrs["transformation_type"])
            vector = [float(v) for v in item.attrs["vector"]]
            if ttype == "translation":
                return
            elif ttype == "rotation":
                if hasattr(value, "__iter__") and len(value):
                    value = value[0]
                if units == "rad":
                    value *= 180 / math.pi
                elif units not in ["deg", "degree", "degrees"]:
                    raise RuntimeError("Invalid units: %s" % units)

                # is the axis moving? Check the values for this axis
                v = item[...]
                if v.min() < v.max():
                    is_scan_axis = True
                else:
                    is_scan_axis = False

                # Is different coordinate system called mcstas
                # Rotate 180 about up if memory serves
                axis_name = item.name.split("/")[-1]
                self._axes.append(vector)
                self._angles.append(float(value))
                self._axis_names.append(str(axis_name))
                self._is_scan_axis.append(is_scan_axis)

            else:
                raise RuntimeError("Unknown transformation_type: %s" % ttype)

        def result(self):
            if self._is_scan_axis.count(True) == 0:
                # XXX not sure how best to handle this, but probably a still so no scan axis
                scan_axis = 0
            else:
                assert self._is_scan_axis.count(True) == 1, (
                    "Only one axis can be a scan axis: %s" % list(self._is_scan_axis)
                )
                scan_axis = flex.first_index(self._is_scan_axis, True)

            # Rotate 180 about up from McStas coordinate system
            cb_op = (-1, 0, 0, 0, 1, 0, 0, 0, -1)
            return cb_op * self._axes, self._angles, self._axis_names, scan_axis

    if vector is None:
        value = nx_file[item][()]
        units = h5str(nx_file[item].attrs["units"])
        ttype = h5str(nx_file[item].attrs["transformation_type"])
        vector = nx_file[item].attrs["vector"]
        if "offset" in nx_file[item].attrs:
            offset = nx_file[item].attrs["offset"]
        else:
            offset = vector * 0.0
        if ttype == "translation":
            value = convert_units(value, units, "mm")
            try:
                vector = vector * value
            except ValueError:
                vector = vector * value[0]
            vector += offset

    visitor = Visitor()
    visitor.visit(nx_file, item)

    visit_dependencies(nx_file, item, visitor.visit)

    return visitor.result()


class NXdetector_module:
    """
    A class to hold a handle to NXdetector_module
    """

    def __init__(self, handle):
        self.handle = handle


class NXdetector_group:
    """
    A class to hold a handle to NXdetector_group
    """

    def __init__(self, handle):
        self.handle = handle


class NXdetector:
    """
    A class to handle a handle to NXdetector
    """

    def __init__(self, handle):
        self.handle = handle

        # Find the NXdetector_modules
        self.modules = [
            NXdetector_module(entry)
            for entry in find_class(self.handle, "NXdetector_module")
        ]

        # Check we've got some stuff
        if not self.modules:
            raise NXValidationError("No NXdetector_module in %s" % self.handle.name)


class NXinstrument:
    """
    A class to hold a handle to NXinstrument
    """

    def __init__(self, handle):
        self.handle = handle

        # Find the NXdetector, any detector groups and the NXbeam
        detectors, detector_groups, beams = find_classes(
            self.handle, "NXdetector", "NXdetector_group", "NXbeam"
        )

        # Check we've got stuff
        if not detectors:
            raise NXValidationError("No NXdetector in %s" % self.handle.name)

        self.detectors = [NXdetector(detector) for detector in detectors]
        self.detector_groups = [NXdetector_group(group) for group in detector_groups]
        self.beams = [NXbeam(beam) for beam in beams]


class NXbeam:
    """
    A class to hold a handle to NXbeam
    """

    def __init__(self, handle):
        self.handle = handle


class NXsample:
    """
    A class to hold a handle to NXsample
    """

    def __init__(self, handle):
        self.handle = handle

        # Find the NXbeam
        self.beams = [NXbeam(beam) for beam in find_class(self.handle, "NXbeam")]


class NXdata:
    """
    A class to hold a handle to NXdata
    """

    def __init__(self, handle):
        self.handle = handle


class NXmxEntry:
    """
    A class to hold a handle to NXmx entries
    """

    def __init__(self, handle):
        self.handle = handle

        # Find the NXinstrument, NXsample & NXdata
        instruments, samples, data_sets = find_classes(
            self.handle, "NXinstrument", "NXsample", "NXdata"
        )

        # Check we've got some stuff
        if not instruments:
            raise NXValidationError("No NXinstrument in %s" % self.handle.name)
        if not samples:
            raise NXValidationError("No NXsample in %s" % self.handle.name)
        if not data_sets:
            raise NXValidationError("No NXdata in %s" % self.handle.name)

        self.instruments = [NXinstrument(instrument) for instrument in instruments]
        self.samples = [NXsample(sample) for sample in samples]
        self.data = [NXdata(data) for data in data_sets]


class NXmxReader:
    """
    A hacky class to read an NXmx file
    """

    def __init__(self, filename=None, handle=None):
        # Get the file handle
        if filename is not None:
            handle = h5py.File(filename, "r", swmr=True)

        # Find the NXmx entries
        self.entries = []
        for entry in find_entries(handle):
            self.entries.append(NXmxEntry(entry))

        # Check we've got some stuff
        if not self.entries:
            raise RuntimeError(
                "Error reading NXmxfile %r. No NXmx entries in file" % filename
            )

    def print_description(self):
        """
        Print a description of the NXmx file

        """
        print(" > Found %d NXmx entries" % len(self.entries))
        for entry in self.entries:
            handle = entry.handle
            instruments = entry.instruments
            samples = entry.samples
            print("  > %s" % handle.name)
            beams = []
            for instrument in instruments:
                handle = instrument.handle
                beams += instrument.beams
                detectors = instrument.detectors
                print("   > %s" % handle.name)
                for detector in detectors:
                    handle = detector.handle
                    modules = detector.modules
                    print("    > %s" % handle.name)
                    for module in modules:
                        handle = module.handle
                        print("     > %s" % handle.name)
            for sample in samples:
                handle = sample.handle
                beams += sample.beams
                print("   > %s" % handle.name)
            for beam in beams:
                handle = beam.handle
                print("    > %s" % handle.name)


def is_nexus_file(filename):
    """
    A hacky function to check if this is a nexus file
    """
    with h5py.File(filename, "r") as handle:
        # Find the NXmx entries
        return bool(find_entries(handle))


class BeamFactory:
    """
    A class to create a beam model from NXmx stuff
    """

    def __init__(self, obj):
        self.obj = obj
        self.model = None
        self.index = None
        self.spectrum = None

    def read_models(self, index=None):
        self.load_model(index)
        return self.model, self.spectrum

    def load_model(self, index=None):
        # Cached model
        if self.model is not None and index == self.index:
            return self.model

        # Get the items from the NXbeam class
        primary_key = "incident_wavelength"
        wavelength = self.obj.handle[primary_key]
        spectrum_wavelengths = wavelength
        spectrum_weights = self.obj.handle.get(primary_key + "_weights")
        if spectrum_weights is None:
            # Handle deprecation: https://github.com/nexusformat/definitions/issues/837
            spectrum_weights = self.obj.handle.get(primary_key + "_weight")

        # If the wavelength array does not represent spectra, look for spectra
        # in the variant chain
        variant_test = wavelength
        has_variant_spectra = False
        while spectrum_weights is None:
            if "variant" in variant_test.attrs:
                variant_key = variant_test.attrs["variant"]
                variant_wavelengths = self.obj.handle[variant_key]
                variant_weights = self.obj.handle.get(variant_key + "_weights")
                if variant_weights is None:
                    # Handle deprecation: https://github.com/nexusformat/definitions/issues/837
                    variant_weights = self.obj.handle.get(variant_key + "_weight")
                if variant_weights is None:
                    variant_test = variant_wavelengths  # Keep looking
                else:
                    # Found spectra
                    spectrum_wavelengths = variant_wavelengths
                    spectrum_weights = variant_weights  # cause while loop to end
                    has_variant_spectra = True
            else:
                break

        if index is None:
            index = 0
        self.index = index

        def get_wavelength(wavelength):
            if wavelength.shape in ((), (1,)):
                wavelength_value = wavelength[()]
            else:
                wavelength_value = wavelength[index]
            wavelength_units = wavelength.attrs["units"]
            wavelength_value = float(
                convert_units(wavelength_value, wavelength_units, "angstrom")
            )
            return wavelength_value

        if spectrum_weights is None:
            # Construct the beam model
            wavelength_value = get_wavelength(wavelength)
            self.model = Beam(direction=(0, 0, 1), wavelength=wavelength_value)
        else:
            self.model = Beam()
            self.model.set_direction((0, 0, 1))

            wavelength_units = spectrum_wavelengths.attrs["units"]

            if len(spectrum_wavelengths.shape) > 1:
                spectrum_wavelengths = spectrum_wavelengths[index]
            else:
                spectrum_wavelengths = spectrum_wavelengths[()]
            if len(spectrum_weights.shape) > 1:
                spectrum_weights = spectrum_weights[index]
            else:
                spectrum_weights = spectrum_weights[()]

            spectrum_wavelengths = convert_units(
                spectrum_wavelengths, wavelength_units, "angstrom"
            )
            spectrum_energies = cctbx.factor_ev_angstrom / spectrum_wavelengths
            self.spectrum = Spectrum(spectrum_energies, spectrum_weights)

            if has_variant_spectra:
                wavelength_value = get_wavelength(wavelength)
                self.model.set_wavelength(wavelength_value)
            else:
                self.model.set_wavelength(self.spectrum.get_weighted_wavelength())
        return self.model


def get_change_of_basis(transformation, include_setting=True):
    """
    Get the 4x4 homogenous coordinate matrix for a given NXtransformation.
    """
    # Change of basis to convert from NeXus to IUCr/ImageCIF convention
    n2i_cob = sqr((-1, 0, 0, 0, 1, 0, 0, 0, -1))

    axis_type = h5str(transformation.attrs["transformation_type"])

    vector = n2i_cob * col(transformation.attrs["vector"]).normalize()
    setting = transformation[0] if include_setting else 0
    units = h5str(transformation.attrs["units"])

    if "offset" in transformation.attrs:
        offset = n2i_cob * col(transformation.attrs["offset"])
        if "offset_units" in transformation.attrs:
            offset_units = transformation.attrs["offset_units"]
        else:
            offset_units = units
        offset = convert_units(offset, offset_units, "mm")
    else:
        offset = col((0, 0, 0))

    # 4x4 change of basis matrix (homogeneous coordinates)
    cob = None

    if axis_type == "rotation":
        if units == "rad":
            deg = False
        elif units in ["deg", "degree", "degrees"]:
            deg = True
        else:
            raise RuntimeError("Invalid units: %s" % units)
        r3 = vector.axis_and_angle_as_r3_rotation_matrix(setting, deg=deg)
        cob = sqr(
            (
                r3[0],
                r3[1],
                r3[2],
                offset[0],
                r3[3],
                r3[4],
                r3[5],
                offset[1],
                r3[6],
                r3[7],
                r3[8],
                offset[2],
                0,
                0,
                0,
                1,
            )
        )
    elif axis_type == "translation":
        setting = convert_units(setting, units, "mm")
        translation = offset + (vector * setting)
        cob = sqr(
            (
                1,
                0,
                0,
                translation[0],
                0,
                1,
                0,
                translation[1],
                0,
                0,
                1,
                translation[2],
                0,
                0,
                0,
                1,
            )
        )
    else:
        raise ValueError("Unrecognized transformation type: %s" % axis_type)

    return cob


def get_depends_on_chain_using_equipment_components(transformation):
    """
    Given an NXtransformation, find the list of dependencies that transformation
    has.  If there are multiple dependencies in a single 'level', as indicated
    by grouping them using equipment_component, then the dependency chain will
    skip the intermediate dependencies, listing only the first at each level.
    """
    chain = [transformation]
    current = transformation

    while True:
        parent_id = h5str(current.attrs["depends_on"])

        if parent_id == ".":
            return chain
        parent = current.parent[parent_id]

        if "equipment_component" in current.attrs:
            eq_comp = current.attrs["equipment_component"]
            parent_eq_comp = parent.attrs["equipment_component"]
            if eq_comp == parent_eq_comp:
                current = parent
                continue
        chain.append(parent)
        current = parent


def get_cumulative_change_of_basis(transformation):
    """Get the 4x4 homogenous coordinate matrix for a given NXtransformation,
    combining it with the change of basis matrices of parent transformations
    with the same equipment component as the given transformation.
    Returns (parent, change of basis matrix), where parent is None if the
    transformation's depends_on is ".".  Parent is the transformation that the
    top level transformation in this chain of transformations depends on.
    """

    cob = get_change_of_basis(transformation)

    parent_id = h5str(transformation.attrs["depends_on"])

    if parent_id == ".":
        return None, cob
    parent = transformation.parent[parent_id]

    if "equipment_component" in transformation.attrs:
        eq_comp = transformation.attrs["equipment_component"]
        parent_eq_comp = parent.attrs["equipment_component"]
        if eq_comp == parent_eq_comp:
            non_matching_parent, parent_cob = get_cumulative_change_of_basis(parent)
            return non_matching_parent, parent_cob * cob

    return parent, cob


class DetectorFactoryFromGroup:
    """
    A class to create a detector model from a NXdetector_group
    """

    def __init__(self, instrument, beam, idx=None):
        assert len(instrument.detector_groups) == 1, "Multiple detectors not supported"

        nx_group = instrument.detector_groups[0].handle
        group_names = nx_group["group_names"]
        group_parent = nx_group["group_parent"]

        # Verify the NXdetector objects specified by the detector group are present
        expected_detectors = []
        root_name = None
        for i, parent_id in enumerate(group_parent):
            assert parent_id in [
                -1,
                1,
            ], (
                "Hierarchy of detectors not supported. Hierarchy of module components within detector elements is supported"
            )

            if parent_id == -1:
                assert root_name is None, "Multiple roots not supported"
                root_name = group_names[i]
            else:
                expected_detectors.append(group_names[i].astype(str))

        assert root_name is not None, "Detector root not found"
        assert sorted(
            os.path.basename(d.handle.name) for d in instrument.detectors
        ) == sorted(expected_detectors), (
            "Mismatch between detector group names and detectors available"
        )

        root = None

        def set_frame(pg, transformation):
            """Function to set the local frame of a panel/panel group given an
            NXtransformation"""
            parent, cob = get_cumulative_change_of_basis(transformation)
            # set up the dxtbx d matrix.  Note use of homogenous coordinates.
            origin = matrix.col((cob * matrix.col((0, 0, 0, 1)))[0:3])
            fast = matrix.col((cob * matrix.col((1, 0, 0, 1)))[0:3]) - origin
            slow = matrix.col((cob * matrix.col((0, 1, 0, 1)))[0:3]) - origin

            pg.set_local_frame(fast.elems, slow.elems, origin.elems)

            name = str(os.path.basename(transformation.name))
            pg.set_name(name)

        self.model = Detector()
        for nx_detector in instrument.detectors:
            # Get the detector type
            if "type" in nx_detector.handle:
                detector_type = str(nx_detector.handle["type"][()])
            else:
                detector_type = "unknown"

            # get the set of leaf modules (handles nested modules)
            modules = [
                module
                for module in nx_detector.modules
                if not find_class(module.handle, "NXdetector_module")
            ]

            # depends_on field for a detector will have NxM entries in it, where
            # N = number of images and M = number of detector modules

            for module_number, nx_detector_module in enumerate(modules):
                # Get the depends_on starting point for this module and for this image
                panel_name = str(os.path.basename(nx_detector_module.handle.name))
                # image size stored slow to fast but dxtbx needs fast to slow
                image_size = tuple(
                    reversed(
                        list(map(int, nx_detector_module.handle["data_size"][-2:]))
                    )
                )

                # Get the trusted range of pixel values - if missing use
                # full range of signed 32 bit int
                try:
                    min_trusted_value = float(nx_detector.handle["underload_value"][()])
                except KeyError:
                    min_trusted_value = -0x7FFFFFFF

                try:
                    max_trusted_value = float(
                        nx_detector.handle["saturation_value"][()]
                    )
                except KeyError:
                    max_trusted_value = 0x7FFFFFFF

                trusted_range = min_trusted_value, max_trusted_value

                fast_pixel_direction_handle = nx_detector_module.handle[
                    "fast_pixel_direction"
                ]
                slow_pixel_direction_handle = nx_detector_module.handle[
                    "slow_pixel_direction"
                ]
                assert (
                    fast_pixel_direction_handle.attrs["depends_on"]
                    == slow_pixel_direction_handle.attrs["depends_on"]
                )
                depends_on = fast_pixel_direction_handle
                fast_pixel_direction_value = convert_units(
                    fast_pixel_direction_handle[0],
                    fast_pixel_direction_handle.attrs["units"],
                    "mm",
                )
                slow_pixel_direction_value = convert_units(
                    slow_pixel_direction_handle[0],
                    slow_pixel_direction_handle.attrs["units"],
                    "mm",
                )
                pixel_size = (
                    float(fast_pixel_direction_value),
                    float(slow_pixel_direction_value),
                )

                # Set up the hierarchical detector by iteraing through the dependencies,
                # starting at the root
                chain = get_depends_on_chain_using_equipment_components(depends_on)
                chain.pop(0)  # Do not want fast/slow pixel directions here
                pg = None
                for transform in reversed(chain):
                    name = str(os.path.basename(transform.name))
                    if (
                        pg is None
                    ):  # The first transform will be the root of the hierarchy
                        if root is None:
                            root = self.model.hierarchy()
                            set_frame(root, transform)
                        else:
                            assert root.get_name() == name, "Found multiple roots"
                        pg = root
                        continue

                    # subsequent times through the loop, pg will be the parent of the
                    # current transform
                    pg_names = [child.get_name() for child in pg]
                    if name in pg_names:
                        pg = pg[pg_names.index(name)]
                    else:
                        pg = pg.add_group()
                        set_frame(pg, transform)

                # pg is now this panel's parent
                p = pg.add_panel()
                fast = fast_pixel_direction_handle.attrs["vector"]
                fast = matrix.col([-fast[0], fast[1], -fast[2]])
                slow = slow_pixel_direction_handle.attrs["vector"]
                slow = matrix.col([-slow[0], slow[1], -slow[2]])
                cob = get_change_of_basis(
                    fast_pixel_direction_handle, include_setting=False
                ) * get_change_of_basis(
                    slow_pixel_direction_handle, include_setting=False
                )
                origin = matrix.col((cob * matrix.col((0, 0, 0, 1)))[0:3])

                p.set_local_frame(fast.elems, slow.elems, origin.elems)

                p.set_name(panel_name)
                p.set_pixel_size(pixel_size)
                p.set_image_size(image_size)
                p.set_type(detector_type)
                p.set_trusted_range(trusted_range)

                if "sensor_thickness" in nx_detector.handle:
                    # Get the detector thickness
                    thickness = nx_detector.handle["sensor_thickness"]
                    thickness_value = float(thickness[()])
                    thickness_units = thickness.attrs["units"]
                    thickness_value = float(
                        convert_units(thickness_value, thickness_units, "mm")
                    )
                    p.set_thickness(thickness_value)

                # Get the detector material
                if "sensor_material" in nx_detector.handle:
                    value = h5str(nx_detector.handle["sensor_material"][()])
                    material = {
                        "Si": "Si",
                        "Silicon": "Si",
                        "Sillicon": "Si",
                        "CdTe": "CdTe",
                        "GaAs": "GaAs",
                    }.get(value)
                    if not material:
                        raise RuntimeError("Unknown material: %s" % value)
                    p.set_material(material)

                    # Compute the attenuation coefficient.
                    # This will fail for undefined composite materials
                    # mu_at_angstrom returns cm^-1, but need mu in mm^-1
                    table = attenuation_coefficient.get_table(material)
                    wavelength = beam.get_wavelength()
                    mu = table.mu_at_angstrom(wavelength) / 10.0
                    p.set_mu(mu)

                if (
                    "sensor_thickness" in nx_detector.handle
                    and "sensor_material" in nx_detector.handle
                ):
                    p.set_px_mm_strategy(
                        ParallaxCorrectedPxMmStrategy(mu, thickness_value)
                    )


class DetectorFactory:
    """
    A class to create a detector model from NXmx stuff
    """

    def __init__(self, obj, beam, shape=None):
        # Get the handles
        nx_file = obj.handle.file
        nx_detector = obj.handle
        nx_module = obj.modules[0].handle

        # Get the detector name and type
        if "type" in nx_detector:
            detector_type = str(nx_detector["type"][()])
        else:
            detector_type = "unknown"
        detector_name = str(nx_detector.name)

        try:
            # underload_value: The lowest value at which pixels for this detector would be reasonably be measured.
            # https://manual.nexusformat.org/classes/applications/NXmx.html#nxmx-entry-instrument-detector-underload-value-field
            min_trusted_value = float(nx_detector["underload_value"][()])
        except KeyError:
            min_trusted_value = -0x7FFFFFFF

        try:
            # saturation_value: The value at which the detector goes into saturation.
            # Data above this value is known to be invalid.
            # https://manual.nexusformat.org/classes/applications/NXmx.html#nxmx-entry-instrument-detector-saturation-value-field
            max_trusted_value = float(nx_detector["saturation_value"][()])
        except KeyError:
            max_trusted_value = 0x7FFFFFFF

        trusted_range = min_trusted_value, max_trusted_value

        # Get the detector thickness
        thickness = nx_detector["sensor_thickness"]
        thickness_value = float(thickness[()])
        thickness_units = thickness.attrs["units"]
        thickness_value = float(convert_units(thickness_value, thickness_units, "mm"))

        # Get the detector material
        material = {
            "Si": "Si",
            "Silicon": "Si",
            "Sillicon": "Si",
            "CdTe": "CdTe",
            "GaAs": "GaAs",
        }.get(h5str(nx_detector["sensor_material"][()]))
        if not material:
            raise RuntimeError(
                "Unknown material: %s" % nx_detector["sensor_material"][()]
            )

        try:
            x_pixel = nx_detector["x_pixel_size"][()] * 1000.0
            y_pixel = nx_detector["y_pixel_size"][()] * 1000.0

            legacy_beam_x = float(x_pixel * nx_detector["beam_center_x"][()])
            legacy_beam_y = float(y_pixel * nx_detector["beam_center_y"][()])
        except KeyError:
            legacy_beam_x = 0
            legacy_beam_y = 0

        # Get the fast pixel size and vector
        fast_pixel_direction = nx_module["fast_pixel_direction"]
        fast_pixel_direction_value = float(fast_pixel_direction[()])
        fast_pixel_direction_units = fast_pixel_direction.attrs["units"]
        fast_pixel_direction_vector = fast_pixel_direction.attrs["vector"]
        fast_pixel_direction_value = convert_units(
            fast_pixel_direction_value, fast_pixel_direction_units, "mm"
        )
        fast_axis = matrix.col(fast_pixel_direction_vector).normalize()

        # Get the slow pixel size and vector
        slow_pixel_direction = nx_module["slow_pixel_direction"]
        slow_pixel_direction_value = float(slow_pixel_direction[()])
        slow_pixel_direction_units = slow_pixel_direction.attrs["units"]
        slow_pixel_direction_vector = slow_pixel_direction.attrs["vector"]
        slow_pixel_direction_value = convert_units(
            slow_pixel_direction_value, slow_pixel_direction_units, "mm"
        )
        slow_axis = matrix.col(slow_pixel_direction_vector).normalize()

        # Get the origin vector - working around if absent
        module_offset = nx_module["module_offset"]
        origin = construct_vector(nx_file, module_offset.name)
        if origin.elems[0] == 0.0 and origin.elems[1] == 0:
            origin = -(origin + legacy_beam_x * fast_axis + legacy_beam_y * slow_axis)

        # Change of basis to convert from NeXus to IUCr/ImageCIF convention
        cob = matrix.sqr((-1, 0, 0, 0, 1, 0, 0, 0, -1))
        origin = cob * matrix.col(origin)
        fast_axis = cob * fast_axis
        slow_axis = cob * slow_axis

        # Ensure that fast and slow axis are orthogonal
        normal = fast_axis.cross(slow_axis)
        slow_axis = -fast_axis.cross(normal)

        # Compute the attenuation coefficient.
        # This will fail for undefined composite materials
        # mu_at_angstrom returns cm^-1, but need mu in mm^-1
        table = attenuation_coefficient.get_table(material)
        wavelength = beam.get_wavelength()
        mu = float(table.mu_at_angstrom(wavelength)) / 10.0

        # Construct the detector model
        pixel_size = (fast_pixel_direction_value, slow_pixel_direction_value)

        # image size stored slow to fast but dxtbx needs fast to slow - assume
        # that the shapes being taken as input are slow -> fast
        if shape:
            image_size = tuple(reversed(shape[-2:]))
        else:
            image_size = tuple(int(x) for x in reversed(nx_module["data_size"][-2:]))

        self.model = Detector()
        self.model.add_panel(
            Panel(
                detector_type,
                detector_name,
                tuple(float(x) for x in fast_axis),
                tuple(float(x) for x in slow_axis),
                tuple(float(x) for x in origin),
                pixel_size,
                image_size,
                trusted_range,
                thickness_value,
                material,
                mu,
            )
        )

        # Set the parallax correction
        for panel in self.model:
            panel.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, thickness_value))
            panel.set_type("SENSOR_PAD")


class GoniometerFactory:
    """
    A class to create a goniometer model from NXmx stuff
    """

    def __init__(self, obj):
        if h5str(obj.handle["depends_on"][()]) == ".":
            self.model = None
        else:
            axes, angles, axis_names, scan_axis = construct_axes(
                obj.handle.file, obj.handle.file[obj.handle["depends_on"][()]].name
            )

            if len(axes) == 1:
                self.model = dxtbx.model.GoniometerFactory.make_goniometer(
                    axes[0], (1, 0, 0, 0, 1, 0, 0, 0, 1)
                )
            else:
                self.model = dxtbx.model.GoniometerFactory.make_multi_axis_goniometer(
                    axes, angles, axis_names, scan_axis
                )


def find_goniometer_rotation(obj):
    if h5str(obj.handle["depends_on"][()]) == ".":
        return
    thing = obj.handle.file[obj.handle["depends_on"][()]]
    tree = get_depends_on_chain_using_equipment_components(thing)
    for t in tree:
        o = obj.handle.file[t.name]
        if h5str(o.attrs["transformation_type"]) == "rotation":
            # if this is changing, assume is scan axis
            v = o[...]
            if v.min() < v.max():
                return o
    raise ValueError("no rotation found")


def find_scanning_axis(obj):
    if h5str(obj.handle["depends_on"][()]) == ".":
        return
    thing = obj.handle.file[obj.handle["depends_on"][()]]
    tree = get_depends_on_chain_using_equipment_components(thing)
    for t in tree:
        o = obj.handle.file[t.name]
        if o[()].size > 1:
            return o


def generate_scan_model(obj, detector_obj):
    """
    Create a scan model from NXmx stuff.
    """
    if h5str(obj.handle["depends_on"][()]) == ".":
        return

    # Get the image and oscillation range - need to search for rotations
    # in dependency tree - if not, find translations or just the thing
    # the sample depends on
    try:
        scan_axis = find_goniometer_rotation(obj)
    except ValueError:
        scan_axis = find_scanning_axis(obj)

    if scan_axis is None:
        scan_axis = obj.handle.file[obj.handle["depends_on"][()]]

    num_images = len(scan_axis)
    image_range = (1, num_images)

    rotn = h5str(scan_axis.attrs["transformation_type"]) == "rotation"

    if num_images > 1 and rotn:
        oscillation = (float(scan_axis[0]), float(scan_axis[1] - scan_axis[0]))
    else:
        # If not a rotation, or only one image, => stills, oscillation range = 0
        angle = float(scan_axis[0]) if rotn else 0
        oscillation = (angle, 0)

    # Get the exposure time
    if "frame_time" in detector_obj.handle:
        frame_time = float(detector_obj.handle["frame_time"][()])
        exposure_time = flex.double(num_images, frame_time)
        epochs = flex.double(num_images)
        for i in range(1, len(epochs)):
            epochs[i] = epochs[i - 1] + exposure_time[i - 1]
    else:
        exposure_time = flex.double(num_images, 0)
        epochs = flex.double(num_images, 0)

    # Construct the model
    return Scan(image_range, oscillation, exposure_time, epochs)


class CrystalFactory:
    """
    A class to create a crystal model from NXmx stuff
    """

    def __init__(self, obj):
        # Get the crystal parameters
        unit_cell_parameters = list(obj.handle["unit_cell"][0])
        unit_cell = cctbx.uctbx.unit_cell(unit_cell_parameters)
        U = list(obj.handle["orientation_matrix"][0].flatten())
        U = matrix.sqr(U)
        B = matrix.sqr(unit_cell.fractionalization_matrix()).transpose()
        A = U * B
        Ai = A.inverse()
        real_space_a = Ai[0:3]
        real_space_b = Ai[3:6]
        real_space_c = Ai[6:9]

        # Get the space group symbol
        space_group_symbol = obj.handle["unit_cell_group"][()]

        # Create the model
        self.model = Crystal(
            real_space_a, real_space_b, real_space_c, space_group_symbol
        )


class DetectorGroupDataList:
    """
    A class to make it easier to access the data from multiple datasets.
    This version brings in all the panels from a detector group with several detectors.
    """

    def __init__(self, datalists):
        self._datalists = datalists
        lengths = [len(datalist) for datalist in datalists]
        self._num_images = lengths[0]
        assert all(length == self._num_images for length in lengths), (
            "Not all datasets are the same length"
        )

    def __len__(self):
        return self._num_images

    def __getitem__(self, index):
        return tuple(itertools.chain.from_iterable(dl[index] for dl in self._datalists))


def get_detector_module_slices(detector):
    """
    Helper function to read data_origin and data_size from the NXdetector_modules in a
    NXdetector.  Returns a list of lists, where each sublist is a list of slices in
    slow to fast order.
    Assumes slices are stored in NeXus in slow to fast order.
    """
    # get the set of leaf modules (handles nested modules)
    modules = []
    for nx_detector_module in detector.modules:
        if not find_class(nx_detector_module.handle, "NXdetector_module"):
            modules.append(nx_detector_module)

    all_slices = []
    for module in modules:
        data_origin = module.handle["data_origin"]
        data_size = module.handle["data_size"]
        all_slices.append(
            [
                slice(int(start), int(start + step), 1)
                for start, step in zip(data_origin, data_size)
            ]
        )
    return all_slices


class MultiPanelDataList:
    """
    A class to make it easier to access the data from multiple datasets.
    Also handles multi-panel data as described in a series of NXdetector_modules
    """

    def __init__(self, datasets, detector):
        self._datasets = datasets
        self._num_images = 0
        self._lookup = []
        self._offset = [0]
        for i, dataset in enumerate(self._datasets):
            self._num_images += dataset.shape[0]
            self._lookup.extend([i] * dataset.shape[0])
            self._offset.append(self._num_images)

        self._all_slices = get_detector_module_slices(detector)

    def __len__(self):
        return self._num_images

    def __getitem__(self, index):
        d = self._lookup[index]
        i = index - self._offset[d]

        all_data = []

        for module_slices in self._all_slices:
            slices = [slice(i, i + 1, 1)]
            slices.extend(module_slices)
            data_as_flex = dataset_as_flex(self._datasets[d], tuple(slices))
            data_as_flex.reshape(
                flex.grid(data_as_flex.all()[-2:])
            )  # handle 3 or 4 dimension arrays
            all_data.append(data_as_flex)
        return tuple(all_data)


DataFactoryCache = collections.namedtuple(
    "DataFactoryCache", "ndim shape filename is_virtual"
)


class DataFactory:
    """
    A class to make it easier to access data from multiple datasets.
    """

    def __init__(self, obj, max_size=0, cached_information=None):
        """
        cached_information is a dictionary of DataFactoryCache named tuples.
        The dictionary key corresponds to the object handle key.
        """

        self.clear_cache()

        DataSetInformation = collections.namedtuple(
            "DataSetInformation", "accessor file shape"
        )
        datasets = []
        for key in sorted(obj.handle):
            if key.startswith("_filename_"):
                continue

            if cached_information and key in cached_information:
                ohk = cached_information[key]
                filename = ohk.filename
            else:
                ohk = obj.handle[key]  # this opens the file
                filename = ohk.file.filename

            # datasets in this context mean ones which contain diffraction images
            # so must have ndim > 1 - for example omega can also be nexus data set
            # but with 1 dimension...
            if ohk.ndim == 1:
                continue

            dsi = DataSetInformation(
                accessor=(lambda obj=obj, key=key: obj.handle[key]),
                file=filename,
                shape=ohk.shape,
            )
            if ohk.is_virtual:
                datasets = [dsi]
                break
            else:
                datasets.append(dsi)

        self._datasets = tuple(datasets)
        self._num_images = 0
        self._lookup = []
        self._offset = [0]
        self._shape = None

        if len(self._datasets) == 1 and max_size:
            self._shape = self._datasets[0].shape
            self._num_images = max_size
            self._lookup.extend([0] * max_size)
            self._offset.append(max_size)
        else:
            for i, dataset in enumerate(self._datasets):
                if self._shape is None:
                    self._shape = dataset.shape
                else:
                    assert self._shape[-2:] == dataset.shape[-2:]
                    self._shape = (self._shape[0] + dataset.shape[0],) + self._shape[1:]
                self._num_images += dataset.shape[0]
                self._lookup.extend([i] * dataset.shape[0])
                self._offset.append(self._num_images)

    def clear_cache(self):
        self._cache = (None, None)

    def __len__(self):
        return self._num_images

    def shape(self):
        return self._shape

    def __getitem__(self, index):
        d = self._lookup[index]
        i = index - self._offset[d]

        # a lock-free most-recently-used file handle cache based on
        # immutability of python tuples
        cached_handle = self._cache
        if self._datasets[d].file == cached_handle[0]:
            data = cached_handle[1]
        else:
            data = self._datasets[d].accessor()
            self._cache = (self._datasets[d].file, data)

        N, height, width = self._datasets[d].shape
        data_as_flex = dataset_as_flex(
            data, (slice(i, i + 1, 1), slice(0, height, 1), slice(0, width, 1))
        )
        data_as_flex.reshape(flex.grid(data_as_flex.all()[1:]))
        return data_as_flex


def detectorgroupdatafactory(obj, instrument):
    """Function to handle reading data from a detector with a NXdetector_group"""

    mapping = {}
    for key in sorted(obj.handle):
        if key.startswith("_filename_"):
            continue

        # When multiple datasets are in NXData, it is required to specify one as the signal
        # In such a case, we skip all datasets except the one flagged as signal
        if "signal" in obj.handle.attrs and key != obj.handle.attrs["signal"]:
            continue

        # datasets in this context mean ones which contain diffraction images
        # so must have ndim > 1 - for example omega can also be nexus data set
        # but with 1 dimension...

        dataset = obj.handle[key]  # this opens the file
        if dataset.ndim == 1:
            continue

        # Map NXdetector names to list of datasets
        dataset_name = key
        found_it = False
        for detector in instrument.detectors:
            if dataset_name in detector.handle:
                found_it = True
                detector_name = os.path.basename(detector.handle.name)
                if detector_name in mapping:
                    assert (
                        dataset_name not in mapping[detector_name]["dataset_names"]
                    ), "Dataset %s found in > 1 NXdetectors" % dataset_name
                    mapping[detector_name]["dataset_names"].add(dataset_name)
                    mapping[detector_name]["datasets"].append(dataset)
                else:
                    mapping[detector_name] = {
                        "dataset_names": {dataset_name},
                        "datasets": [dataset],
                        "detector": detector,
                    }
        assert found_it, "Couldn't match dataset %s to a NXdetector" % dataset_name

    # Create a list of multipanel datalist objects
    return DetectorGroupDataList(
        [
            MultiPanelDataList(
                detector_mapping["datasets"], detector_mapping["detector"]
            )
            for detector_mapping in mapping.values()
        ]
    )


class MaskFactory:
    """
    A class to create an object to hold the pixel mask data
    """

    def __init__(self, objects, index=None):
        def make_mask(dset, index):
            i = 0 if index is None else index
            mask = []
            for module_slices in all_slices:
                assert len(dset.shape) in [len(module_slices), len(module_slices) + 1]
                if len(dset.shape) == len(module_slices):
                    slices = []  # single image mask
                else:
                    slices = [slice(i, i + 1, 1)]  # multi-image mask
                slices.extend(module_slices)
                data_as_flex = dataset_as_flex_int(dset.id.id, tuple(slices))
                data_as_flex.reshape(
                    flex.grid(data_as_flex.all()[-2:])
                )  # handle 3 or 4 dimension arrays
                mask.append(data_as_flex == 0)
            return tuple(mask)

        self.mask = None
        for obj in objects:
            handle = obj.handle
            if "pixel_mask_applied" in handle and handle["pixel_mask_applied"]:
                if self.mask is None:
                    self.mask = []
                if "pixel_mask" in handle:
                    shape = handle["pixel_mask"].shape
                    all_slices = get_detector_module_slices(obj)
                    if len(all_slices) == 1:
                        all_slices = [[slice(0, shape[0], 1), slice(0, shape[1], 1)]]
                    self.mask.extend(list(make_mask(handle["pixel_mask"], index)))
                elif "detectorSpecific" in handle:
                    if "pixel_mask" in handle["detectorSpecific"]:
                        shape = handle["detectorSpecific"]["pixel_mask"].shape
                        all_slices = get_detector_module_slices(obj)
                        if len(all_slices) == 1:
                            all_slices = [
                                [slice(0, shape[0], 1), slice(0, shape[1], 1)]
                            ]
                        self.mask.extend(
                            list(
                                make_mask(
                                    handle["detectorSpecific"]["pixel_mask"], index
                                )
                            )
                        )
        if self.mask is not None:
            self.mask = tuple(self.mask)



"""Reads a multi-tile CBF image, discovering its detector geometry automatically"""

from __future__ import annotations

import sys

import numpy
import pycbf

from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.format.FormatCBF import FormatCBF
from dxtbx.format.FormatCBFFull import FormatCBFFull
from dxtbx.format.FormatStill import FormatStill
from dxtbx.model.detector import Detector
from dxtbx.model.detector_helpers import find_undefined_value, find_underload_value


def angle_and_axis(basis):
    """Normalize a quaternion and return the angle and axis
    @param params metrology object"""
    q = matrix.col(basis.orientation).normalize()
    return q.unit_quaternion_as_axis_and_angle(deg=True)


class cbf_wrapper(pycbf.cbf_handle_struct):
    """Wrapper class that provides convenience functions for working with cbflib"""

    def add_category(self, name, columns):
        """Create a new category and populate it with column names"""
        self.new_category(name.encode())
        for column in columns:
            self.new_column(column.encode())

    def add_row(self, data):
        """Add a row to the current category.  If data contains more entries than
        there are columns in this category, then the remainder is truncated
        Use '.' for an empty value in a row."""
        self.new_row()
        self.rewind_column()
        for item in data:
            try:
                self.set_value(item.encode())
            except AttributeError:
                self.set_value(item)
            if item == ".":
                self.set_typeofvalue(b"null")
            try:
                self.next_column()
            except Exception:
                break

    def has_sections(self):
        """True if the cbf has the array_structure_list_section table, which
        changes how its data is stored in the binary sections
        """
        try:
            self.find_category(b"array_structure_list_section")
            return True
        except Exception as e:
            if "CBF_NOTFOUND" in str(e):
                return False
            raise e

    def add_frame_shift(self, basis, axis_settings):
        """Add an axis representing a frame shift (a rotation axis with an offset)"""
        angle, axis = angle_and_axis(basis)

        if angle == 0:
            axis = (0, 0, 1)

        if basis.include_translation:
            translation = basis.translation
        else:
            translation = (0, 0, 0)

        self.add_row(
            [
                basis.axis_name,
                "rotation",
                "detector",
                basis.depends_on,
                str(axis[0]),
                str(axis[1]),
                str(axis[2]),
                str(translation[0]),
                str(translation[1]),
                str(translation[2]),
                basis.equipment_component,
            ]
        )

        axis_settings.append([basis.axis_name, "FRAME1", str(angle), "0"])


class FormatCBFMultiTile(FormatCBFFull):
    """An image reading class multi-tile CBF files"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        try:
            cbf_handle = pycbf.cbf_handle_struct()
            cbf_handle.read_widefile(image_file.encode(), pycbf.MSG_DIGEST)
        except Exception as e:
            if "CBFlib Error" in str(e):
                return False

        # check if multiple arrays
        try:
            return cbf_handle.count_elements() > 1
        except Exception as e:
            if "CBFlib Error" in str(e):
                return False

    def _start(self):
        """Open the image file as a cbf file handle, and keep this somewhere
        safe."""
        FormatCBF._start(self)  # Note, skip up an inheritance level

    def detectorbase_start(self):
        pass

    def _get_cbf_handle(self):
        try:
            return self._cbf_handle
        except AttributeError:
            self._cbf_handle = cbf_wrapper()
            self._cbf_handle.read_widefile(self._image_file.encode(), pycbf.MSG_DIGEST)
            return self._cbf_handle

    def _detector(self):
        """Return a working detector instance."""

        cbf = self._get_cbf_handle()

        d = Detector()

        for i in range(cbf.count_elements()):
            ele_id = cbf.get_element_id(i)
            cbf.find_category(b"diffrn_data_frame")
            cbf.find_column(b"detector_element_id")
            cbf.find_row(ele_id)
            cbf.find_column(b"array_id")
            array_id = cbf.get_value()

            cbf_detector = cbf.construct_detector(i)

            p = d.add_panel()
            p.set_name(array_id)

            # code adapted below from dxtbx.model.detector.DetectorFactory.imgCIF_H
            pixel = (
                cbf_detector.get_inferred_pixel_size(1),
                cbf_detector.get_inferred_pixel_size(2),
            )

            fast = cbf_detector.get_detector_axes()[0:3]
            slow = cbf_detector.get_detector_axes()[3:6]
            origin = cbf_detector.get_pixel_coordinates_fs(0, 0)

            size = tuple(reversed(cbf.get_image_size(0)))

            try:
                min_trusted_value = find_underload_value(cbf)
            except Exception:
                try:
                    # By convention, if underload is not set, then assume the minimum
                    # trusted pixel is 1 more than the undefined pixel
                    min_trusted_value = find_undefined_value(cbf) + 1
                except Exception:
                    min_trusted_value = 0
            try:
                max_trusted_value = cbf.get_overload(i) - 1
            except Exception:
                max_trusted_value = 1.0e6

            trusted_range = (min_trusted_value, max_trusted_value)

            try:
                cbf.find_column(b"gain")
                cbf.select_row(i)
                gain = cbf.get_doublevalue()
            except Exception as e:
                if "CBF_NOTFOUND" not in str(e):
                    raise
                gain = 1.0

            cbf_detector.__swig_destroy__(cbf_detector)
            del cbf_detector

            p.set_local_frame(fast, slow, origin)

            p.set_pixel_size(tuple(map(float, pixel)))
            p.set_image_size(size)
            p.set_trusted_range(tuple(map(float, trusted_range)))
            p.set_gain(gain)
            # p.set_px_mm_strategy(px_mm) FIXME

        return d

    def _beam(self):
        """Return a working beam instance."""

        return self._beam_factory.imgCIF_H(self._get_cbf_handle())

    def get_raw_data(self):
        if self._raw_data is None:
            self._raw_data = []

            cbf = self._get_cbf_handle()

            # find the data
            cbf.select_category(0)
            while cbf.category_name().lower() != "array_data":
                try:
                    cbf.next_category()
                except Exception:
                    return None
            cbf.select_column(0)
            cbf.select_row(0)

            d = self.get_detector()

            for panel in d:
                name = panel.get_name()
                cbf.find_column(b"array_id")
                assert name == cbf.get_value()

                cbf.find_column(b"data")
                assert cbf.get_typeofvalue().find(b"bnry") > -1

                image_string = cbf.get_realarray_as_string()
                image = flex.double(numpy.frombuffer(image_string, numpy.float))

                parameters = cbf.get_realarrayparameters_wdims_fs()
                image_size = (parameters[6], parameters[5])

                image.reshape(flex.grid(*image_size))

                self._raw_data.append(image)

                try:
                    cbf.next_row()
                except Exception:
                    break
            assert len(d) == len(self._raw_data)

        return tuple(self._raw_data)


class FormatCBFMultiTileStill(FormatStill, FormatCBFMultiTile):
    """An image reading class for full CBF format images i.e. those from
    a variety of cameras which support this format. Custom derived from
    the FormatStill to handle images without a gonimeter or scan"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        header = FormatCBFMultiTile.get_cbf_header(image_file)

        # According to ImageCIF, "Data items in the DIFFRN_MEASUREMENT_AXIS
        # category associate axes with goniometers."
        # http://www.iucr.org/__data/iucr/cifdic_html/2/cif_img.dic/Cdiffrn_measurement_axis.html
        if "diffrn_measurement_axis" in header:
            return False
        return True


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMultiTile.understand(arg))



"""An ImageFormat class to read MarIP-format image"""

from __future__ import annotations

import sys

from iotbx.detectors.detectorbase import DetectorImageBase

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format

try:
    # Try to import the modernised pycbf
    import pycbf  # noqa: F401
    from pycbf.img import Img

    import scitbx.array_family.flex as flex
except ModuleNotFoundError:
    Img: type | None = None  # type: ignore
    # Replicate fallback logic from iotbx.detectors.marIP
    try:
        from cbflib_adaptbx import Mar345Adaptor
    except ModuleNotFoundError:
        from iotbx.detectors.marIP import NullAdaptor as Mar345Adaptor

if Img is not None:
    # We have the modernised pycbf with the img bindings.
    # Declare Mar345Adaptor in pure python
    class Mar345Adaptor:  # type: ignore # noqa: F811
        def __init__(self, filename):
            self.filename = filename

            self._read_header = False
            self._read_data = False
            self._data_position = None
            self._metadata = None
            self._img = Img()
            self._img.set_tags(0)
            self._img.set_dimensions(0, 0)

        def read_header(self):
            if self._read_header:
                return
            with open(self.filename, "rb") as f:
                self._metadata = self._img.read_mar345header(f)
                self._data_position = f.tell()
            detector = self._img.get_field("DETECTOR")
            if "mar" not in detector.lower() or "345" not in detector.lower():
                raise RuntimeError(
                    f"Detector type other than mar345 from mar345 reader ({detector})"
                )

        def read_data(self):
            if self._read_data:
                return
            self.read_header()
            with open(self.filename, "rb") as f:
                f.seek(self._data_position)
                self._img.read_mar345data(f, self._metadata)

        def rawdata(self):
            self.read_data()
            return flex.int(self._img.image)

        def columns(self):
            return self._img.columns

        def rows(self):
            return self._img.rows

        def size1(self):
            self.read_header()
            return self._metadata[0]

        def size2(self):
            self.read_header()
            return self._metadata[2]

        def pixel_size(self):
            self.read_header()
            return self._img.get_number("PIXEL SIZE")

        def wavelength(self):
            self.read_header()
            return self._img.get_number("WAVELENGTH")

        def distance(self):
            self.read_header()
            return self._img.get_number("DISTANCE")

        def gain(self):
            return 1.55

        def overload(self):
            # Determined from a single image.
            return 249862

        def osc_range(self):
            self.read_header()
            return self._img.get_number("OSCILLATION RANGE")

        def osc_start(self):
            self.read_header()
            return self._img.get_number("PHI")

        def twotheta(self):
            self.read_header()
            return self._img.get_number("TWOTHETA")

        def exposure_time(self):
            self.read_header()
            return self._img.get_number("EXPOSURE_TIME")


class MARIPImage(DetectorImageBase):
    def __init__(self, filename):
        DetectorImageBase.__init__(self, filename)
        self.adaptor = Mar345Adaptor(filename)
        self.vendortype = "MARIP"

    def readHeader(self):
        self.parameters = {
            "SIZE1": self.adaptor.size1(),
            "SIZE2": self.adaptor.size2(),
            "CCD_IMAGE_SATURATION": self.adaptor.overload(),
            "PIXEL_SIZE": self.adaptor.pixel_size(),
            "OSC_START": self.adaptor.osc_start(),
            "DISTANCE": self.adaptor.distance(),
            "WAVELENGTH": self.adaptor.wavelength(),
            "BEAM_CENTER_X": self.beam_center_slow(),
            "BEAM_CENTER_Y": self.beam_center_fast(),
            "OSC_RANGE": self.adaptor.osc_range(),
            "TWOTHETA": self.adaptor.twotheta(),
            "DETECTOR_SN": 0,
        }

    def beam_center_slow(self):
        return self.adaptor.size1() * self.adaptor.pixel_size() / 2.0

    def beam_center_fast(self):
        return self.adaptor.size2() * self.adaptor.pixel_size() / 2.0

    def fileLength(self):
        return 0

    def getEndian(self):
        return 0

    def read(self):
        self.bin_safe_set_data(self.adaptor.rawdata())

    def dataoffset(self):
        return 0

    def integerdepth(self):
        return 0


class FormatMarIP(Format):
    """An image reading class for MarIP-format images
    Positive identification:  first 140 bytes contain the string "mar research"
    """

    @staticmethod
    def understand(image_file):
        try:
            with FormatMarIP.open_file(image_file, "rb") as fh:
                return b"mar research" in fh.read(140)
        except OSError:
            return False

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)

        Format.__init__(self, image_file, **kwargs)

    def detectorbase_start(self):
        pass

    def _start(self):
        """Open the image file, read the image header, copy the key / value
        pairs into an internal dictionary self._header_dictionary along with
        the length of the header in bytes self._header_size."""
        self.detectorbase = MARIPImage(self._image_file)
        self.detectorbase.readHeader()

    def _goniometer(self):
        return self._goniometer_factory.single_axis()

    def _detector(self):
        """Return a model for a simple detector, which at the moment insists
        that the offsets and rotations are all 0.0."""

        assert self.detectorbase.parameters["TWOTHETA"] == 0.0

        return self._detector_factory.simple(
            sensor="IMAGE_PLATE",
            distance=self.detectorbase.parameters["DISTANCE"],
            beam_centre=(
                self.detectorbase.parameters["BEAM_CENTER_X"],
                self.detectorbase.parameters["BEAM_CENTER_Y"],
            ),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(
                self.detectorbase.parameters["PIXEL_SIZE"],
                self.detectorbase.parameters["PIXEL_SIZE"],
            ),
            image_size=(
                self.detectorbase.parameters["SIZE1"],
                self.detectorbase.parameters["SIZE2"],
            ),
            trusted_range=(0, self.detectorbase.parameters["CCD_IMAGE_SATURATION"]),
            mask=[],
        )

    def _beam(self):
        """Return a simple model for the beam."""

        return self._beam_factory.simple(self.detectorbase.parameters["WAVELENGTH"])

    def _scan(self):
        """Return the scan information for this image."""

        return self._scan_factory.single_file(
            filename=self._image_file,
            exposure_times=self.detectorbase.adaptor.exposure_time(),
            osc_start=self.detectorbase.parameters["OSC_START"],
            osc_width=self.detectorbase.parameters["OSC_RANGE"],
            epoch=None,
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatMarIP.understand(arg))



from __future__ import annotations

import sys

from iotbx.detectors.macscience import DIPImage

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format


class FormatDIP2030b(Format):
    @staticmethod
    def understand(image_file):
        # for MacScience DIP2030b only, file size is exactly 18001024 bytes
        headerstart = 3000 * 3000 * 2
        try:
            with FormatDIP2030b.open_file(image_file, "rb") as fh:
                fh.seek(headerstart)
                rawheader = fh.read(1024)
                eof = fh.read(1)  # end of file
        except OSError:
            return False

        return eof == b"" and rawheader[0:3] == b"DIP"

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        super().__init__(image_file, **kwargs)

    def detectorbase_start(self):
        pass

    def _start(self):
        self.detectorbase = DIPImage(self._image_file)
        self.detectorbase.readHeader()

    def _goniometer(self):
        return self._goniometer_factory.single_axis()

    def _detector(self):
        """Return a model for a simple detector"""

        # twotheta = self.detectorbase.parameters["TWOTHETA"]
        # At present, ignore non-zero two theta for the dxtbx model
        # XXX Return to this issue later.
        return self._detector_factory.simple(
            sensor="IMAGE_PLATE",
            distance=self.detectorbase.distance,
            beam_centre=(self.detectorbase.beamx, self.detectorbase.beamy),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(self.detectorbase.pixel_size, self.detectorbase.pixel_size),
            image_size=(self.detectorbase.size1, self.detectorbase.size2),
            trusted_range=(0, self.detectorbase.saturation),
            mask=[],
        )  # a list of dead rectangles

    def _beam(self):
        """Return a simple model for the beam."""

        return self._beam_factory.simple(self.detectorbase.wavelength)

    def _scan(self):
        """Return the scan information for this image."""

        return self._scan_factory.single_file(
            filename=self._image_file,
            exposure_times=self.detectorbase.parameters["TIME"],
            osc_start=self.detectorbase.osc_start,
            osc_width=self.detectorbase.deltaphi,
            epoch=None,
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatDIP2030b.understand(arg))



from __future__ import annotations

import sys

import h5py

from scitbx.array_family import flex

from dxtbx import flumpy
from dxtbx.format.FormatHDF5 import FormatHDF5


class FormatHDF5ESRFJungfrau4M(FormatHDF5):
    # A class to understand still-shot images from ESRF collected on a Jungfrau 4M.

    _cached_mask = None

    @staticmethod
    def understand(image_file):
        with h5py.File(image_file, "r") as h5_handle:
            if len(h5_handle) != 1:
                return False
            key = list(h5_handle.keys())[0]
            if not isinstance(
                h5_handle[key], h5py.Group
            ):  # check we're not trying to read a 'dataset' object
                return False
            if "instrument" not in h5_handle[key]:
                return False
            # instrument name is of form jungfrau4m_rr[X]_smx where X is empty, 4 or another number
            instrument = list(h5_handle[key]["instrument"].keys())[0]
            if not instrument.startswith("jungfrau4m_rr"):
                return False
            if not instrument.endswith("smx"):
                return False
        return True

    def _start(self):
        super()._start()
        image_file = self.get_image_file()
        self._h5_handle = h5py.File(image_file, "r")
        self.key = list(self._h5_handle.keys())[0]
        self.instrument_name = list(self._h5_handle[self.key]["instrument"].keys())[0]
        instrument = self._h5_handle[self.key]["instrument"][self.instrument_name]
        self.n_images = instrument["data"].shape[0]
        self.adus_per_photon = instrument["detector_information"]["adus_per_photon"]
        self.image_size = tuple(instrument["data"].shape[1:])
        wavelength = instrument["beam"]["incident_wavelength"][()]
        x_pixel_size = (
            instrument["detector_information"]["x_pixel_size"][()] * 1000
        )  # convert m to mm
        y_pixel_size = (
            instrument["detector_information"]["y_pixel_size"][()] * 1000
        )  # convert m to mm
        distance = (
            instrument["detector_information"]["detector_distance"][()] * 1000
        )  # convert m to mm
        beam_center_x = instrument["detector_information"]["beam_center_x"][()]  # in px
        beam_center_y = instrument["detector_information"]["beam_center_y"][()]  # in px

        beam_center_x *= x_pixel_size
        beam_center_y *= y_pixel_size
        trusted_range = (
            instrument["detector_information"]["underload_value"][()],
            instrument["detector_information"]["saturation_value"][()],
        )
        exposure_time = instrument["acquisition"]["exposure_time"][()]

        self._detector_model = self._detector_factory.simple(
            sensor="UNKNOWN",
            distance=distance,
            beam_centre=(
                beam_center_x,
                beam_center_y,
            ),
            fast_direction="+x",
            slow_direction="-y",
            pixel_size=(
                x_pixel_size,
                y_pixel_size,
            ),
            image_size=(self.image_size[1], self.image_size[0]),
            trusted_range=trusted_range,
            mask=self.get_static_mask(),
        )
        self._beam_model = self._beam_factory.simple(wavelength)
        self._scan_model = self._scan_factory.make_scan(
            image_range=(1, self.n_images),
            exposure_times=exposure_time,
            oscillation=(0.0, 0.0),
            epochs=list(range(self.n_images)),
        )
        # Add a placeholder goniometer model, which has no practical effect on processing as the oscillation is 0.
        # Some dxtbx format logic assumes both or neither scan + goniometer are None for still images
        self._goniometer_model = self._goniometer_factory.known_axis((0, 1, 0))

    def get_raw_data(self, index=None):
        if index is None:
            index = 0
        # data can be int32 with adus_per_photon != 1.0 or float16 with adus_per_photon == 1.0
        data = (
            (
                self._h5_handle[self.key]["measurement"]["data"][index]
                / self.adus_per_photon
            )
            if self.adus_per_photon != 1.0
            else self._h5_handle[self.key]["measurement"]["data"][index]
        )
        return flex.double(data.astype(float))

    def get_num_images(self):
        return self.n_images

    def get_beam(self, index=None):
        return self._beam(index)

    def _beam(self, index=None):
        return self._beam_model

    def get_detector(self, index=None):
        return self._detector(index)

    def get_static_mask(self):
        if FormatHDF5ESRFJungfrau4M._cached_mask is None:
            mask = self._h5_handle[self.key]["instrument"][self.instrument_name][
                "detector_information"
            ]["pixel_mask"]
            mask = flumpy.from_numpy(mask[()])
            mask_array = mask == 0
            FormatHDF5ESRFJungfrau4M._cached_mask = mask_array
        return FormatHDF5ESRFJungfrau4M._cached_mask

    def _detector(self, index=None):
        return self._detector_model

    def _goniometer(self):
        return self._goniometer_model

    def _scan(self):
        return self._scan_model


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatHDF5ESRFJungfrau4M.understand(arg))



"""Support for the Rigaku Oxford Diffraction image format

NB: Rigaku datasets may use a non-zero-padded image incremental serial number.
At present, this is not compatible with assumptions in dxtbx. In order to
import these datasets, the images should be renumbered first.

See https://github.com/cctbx/dxtbx/issues/646 for details."""

# Takanori Nakane took David Waterman's code to parse headers from
#  https://github.com/cctbx/cctbx_project/commit/b95467f3b2a70a37eeb820ea294128a32551700c
# and heavily modified it. The original commit in the cctbx_project repository is orphan now.
from __future__ import annotations

__author__ = "David Waterman, Takanori Nakane"
__copyright__ = "Copyright 2018-2023 United Kingdom Research and Innovation & 2022-2023 Takanori Nakane"
__license__ = "BSD 3-clause"

import re
import struct

import numpy as np

from scitbx import matrix
from scitbx.array_family import flex
from scitbx.math import r3_rotation_axis_and_angle_as_matrix

from dxtbx.ext import uncompress_rod_TY5_TY6
from dxtbx.format.Format import Format
from dxtbx.model.beam import Probe
from dxtbx.model.detector import Detector


class FormatROD(Format):
    """An image reading class for Rigaku Oxford Diffraction images.

    The Rigaku Oxford Diffraction format is used by CrysAlis(Pro).
    """

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        from dxtbx import IncorrectFormatError

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)

        Format.__init__(self, image_file, **kwargs)

    @staticmethod
    def understand(image_file):
        with FormatROD.open_file(image_file, "rb") as f:
            try:
                hdr = f.read(256).decode("ascii")
            except UnicodeDecodeError:
                return False

        lines = hdr.splitlines()
        if len(lines) < 2:
            return False

        vers = lines[0].split()
        if len(vers) < 2 or vers[0] != "OD" or vers[1] != "SAPPHIRE":
            return False

        compression = lines[1].split("=")
        if compression[0] != "COMPRESSION":
            return False

        return True

    @staticmethod
    def _read_ascii_header(image_file):
        """Read the ASCII header comprising the first 256 bytes of the file"""

        hd = {}
        with FormatROD.open_file(image_file, "rb") as f:
            hdr = f.read(256).decode("ascii")
        lines = hdr.splitlines()

        vers = lines[0].split()
        if len(vers) < 2 or vers[0] != "OD" or vers[1] != "SAPPHIRE":
            raise ValueError("Wrong header format")
        hd["version"] = float(vers[-1])
        if hd["version"] != 4.0:
            raise NotImplementedError(
                f"FormatROD: only header version 4.0 is supported but got {hd['version']}"
            )

        compression = lines[1].split("=")
        if compression[0] != "COMPRESSION":
            raise ValueError("Wrong header format")
        hd["compression"] = compression[1]

        # Extract definitions from the 3rd - 5th line
        defn = re.compile(r"([A-Z]+=[ 0-9]+)")
        for line in lines[2:5]:
            sizes = defn.findall(line)
            for s in sizes:
                n, v = s.split("=")
                hd[n] = int(v)

        hd["time"] = lines[5].split("TIME=")[-1].strip("\x1a").rstrip()

        return hd

    @staticmethod
    def _read_binary_header(
        image_file,
        offset=256,
        nbytes=5120,
        general_nbytes=512,
        special_nbytes=768,
        km4gonio_nbytes=1024,
        statistics_nbytes=512,
        history_nbytes=2048,
    ):
        """Read the most relevant parameters from the binary header"""

        with FormatROD.open_file(image_file, "rb") as f:
            # General section
            f.seek(offset)
            bin_x, bin_y = struct.unpack("<hh", f.read(4))
            f.seek(offset + 22)
            chip_npx_x, chip_npx_y, im_npx_x, im_npx_y = struct.unpack(
                "<hhhh", f.read(8)
            )
            f.seek(offset + 36)
            num_points = struct.unpack("<I", f.read(4))[0]
            num_image = struct.unpack("<h", f.read(2))[0]
            if num_points != im_npx_x * im_npx_y:
                raise ValueError("Cannot interpret binary header")
            if num_image != 1:
                raise NotImplementedError("num_image != 1 is not supported.")

            # Special section
            f.seek(offset + general_nbytes + 56)
            gain = struct.unpack("<d", f.read(8))[0]
            f.seek(offset + general_nbytes + 464)
            overflow_flag, overflow_after_remeasure_flag = struct.unpack(
                "<hh", f.read(4)
            )
            f.seek(offset + general_nbytes + 472)
            overflow_threshold = struct.unpack("<l", f.read(4))[0]
            f.seek(offset + general_nbytes + 480)
            exposure_time_sec, overflow_time_sec = struct.unpack("<dd", f.read(16))
            f.seek(offset + general_nbytes + 548)
            detector_type = struct.unpack("<l", f.read(4))[0]
            f.seek(offset + general_nbytes + 568)
            real_px_size_x, real_px_size_y = struct.unpack("<dd", f.read(16))

            # Goniometer section
            f.seek(offset + general_nbytes + special_nbytes + 284)
            # angles for OMEGA, THETA, CHI(=KAPPA), PHI,
            # OMEGA_PRIME (also called DETECTOR_AXIS; what's this?), THETA_PRIME
            start_angles_steps = struct.unpack("<llllllllll", f.read(40))
            end_angles_steps = struct.unpack("<llllllllll", f.read(40))
            f.seek(offset + general_nbytes + special_nbytes + 368)
            step_to_rad = struct.unpack("<dddddddddd", f.read(80))
            f.seek(offset + general_nbytes + special_nbytes + 552)
            # FIXME: I don't know what these are. Isn't the beam along e1 by definition??
            beam_rotn_around_e2, beam_rotn_around_e3 = struct.unpack("<dd", f.read(16))
            alpha1_wavelength = struct.unpack("<d", f.read(8))[0]
            alpha2_wavelength = struct.unpack("<d", f.read(8))[0]
            alpha12_wavelength = struct.unpack("<d", f.read(8))[0]
            f.seek(offset + general_nbytes + special_nbytes + 640)
            # detector rotation in degrees along e1, e2, e3
            detector_rotns = struct.unpack("<ddd", f.read(24))
            # direct beam position when all angles are zero (FIXME: not completely sure)
            origin_px_x, origin_px_y = struct.unpack("<dd", f.read(16))
            # alpha and beta are angles between KAPPA(=CHI) and THETA, and e3.
            angles_in_deg = struct.unpack(
                "<dddd", f.read(32)
            )  # alpha, beta, gamma, delta
            f.seek(offset + general_nbytes + special_nbytes + 712)
            distance_mm = struct.unpack("<d", f.read(8))[0]

        return {
            "bin_x": bin_x,
            "bin_y": bin_y,
            "chip_npx_x": chip_npx_x,
            "chip_npx_y": im_npx_y,
            "im_npx_x": im_npx_x,
            "im_npx_y": im_npx_y,
            "gain": gain,
            "overflow_flag": overflow_flag,
            "overflow_after_remeasure_flag": overflow_after_remeasure_flag,
            "overflow_threshold": overflow_threshold,
            "exposure_time_sec": exposure_time_sec,
            "overflow_time_sec": overflow_time_sec,
            "detector_type": detector_type,
            "real_px_size_x": real_px_size_x,
            "real_px_size_y": real_px_size_y,
            "start_angles_steps": start_angles_steps,
            "end_angles_steps": end_angles_steps,
            "step_to_rad": step_to_rad,
            "beam_rotn_around_e2": beam_rotn_around_e2,
            "beam_rotn_around_e3": beam_rotn_around_e3,
            "alpha1_wavelength": alpha1_wavelength,
            "alpha2_wavelength": alpha2_wavelength,
            "alpha12_wavelength": alpha12_wavelength,
            "detector_rotns": detector_rotns,
            "origin_px_x": origin_px_x,
            "origin_px_y": origin_px_y,
            "angles_in_deg": angles_in_deg,
            "distance_mm": distance_mm,
        }

    def _start(self):
        """Open the image file, read useful metadata into an internal dictionary
        self._header_dictionary"""

        self._txt_header = self._read_ascii_header(self._image_file)
        # TODO: the block sizes in the binary header should be taken from the
        # ASCII header.
        self._bin_header = self._read_binary_header(self._image_file)

        self._gonio_start_angles = (
            np.array(self._bin_header["start_angles_steps"])
            * np.nan_to_num(np.array(self._bin_header["step_to_rad"]))
            / np.pi
            * 180
        )
        self._gonio_end_angles = (
            np.array(self._bin_header["end_angles_steps"])
            * np.nan_to_num(np.array(self._bin_header["step_to_rad"]))
            / np.pi
            * 180
        )

        self._scan_axis = -1
        for axis in [0, 3]:  # 0 - OMEGA, 3 - PHI; the default is omega scan
            if self._gonio_start_angles[axis] != self._gonio_end_angles[axis]:
                self._scan_axis = axis
                break

        return

    # Rigaku/Oxford coordinate system:
    # - e3: parallel to OMEGA (and THETA), upwards
    # - e1: along the beam, from the crystal to the source
    # - e2: completes the right handed coordinate system
    #
    # The detector origin is at the lower left corner looking from the crystal (or the source).
    # - slow axis: (roughly) vertical and towards the ceiling
    # - fast axis: (roughly) horizontal and towards right (at THETA = 0)
    #
    # Importantly, positive THETA, OMEGA, PHI rotations are CLOCKWISE looking down from the ceiling,
    # while positive CHI=KAPPA rotation is CLOCKWISE looking from the detector at OMEGA=0, THETA=0.
    # See specifications uploaded in https://github.com/cctbx/dxtbx/issues/11#issue-434741559.
    #
    # In this dxtbx coordinate system:
    # - Z: along the beam, from the crystal to the source
    # - Y: vertical, towards the ceiling
    # - X: completes the right handed coordinate system
    # Thus, Z = e1, Y = e3, X = e2.
    #
    # The coordinate system of XDS.INP generated by CrysAlisPro:
    # - Z: along the beam, from the source to the crystal
    # - Y: vertical, towards the ground
    # - X: completes the right handed coordinate system
    # Thus, Y and Z are pointing the opposite directions from DIALS.

    def get_goniometer(self, index=None):
        return Format.get_goniometer(self)

    def _goniometer(self):
        if self._scan_axis == -1:
            raise NotImplementedError("Still shots not implemented yet.")
        elif self._scan_axis == 0:  # OMEGA
            dxtbx_scan_axis = 2
        elif self._scan_axis == 3:  # PHI
            dxtbx_scan_axis = 0
        else:
            pass  # should not happen

        # FIXME: sometimes XDS.INP generated by CrysAlisPro has
        #  tiny deviations from (0, 1, 0). I don't know how to calculate it.
        alpha_rad = self._bin_header["angles_in_deg"][0] * np.pi / 180
        axes = flex.vec3_double(
            ((0, -1, 0), (0, -np.cos(alpha_rad), np.sin(alpha_rad)), (0, -1, 0))
        )

        # angles[self._scan_axis] is not used anyway
        # angles are in degrees!
        angles = flex.double(
            (
                self._gonio_start_angles[3],
                self._gonio_start_angles[2],
                self._gonio_start_angles[0],
            )
        )
        names = flex.std_string(("PHI", "KAPPA=CHI", "OMEGA"))

        return self._goniometer_factory.make_multi_axis_goniometer(
            axes, angles, names, dxtbx_scan_axis
        )

    def get_beam(self, index=None):
        return Format.get_beam(self)

    def _beam(self):
        wavelength = self._bin_header["alpha12_wavelength"]
        if wavelength <= 0.05:
            probe = Probe.electron
        else:
            probe = Probe.xray

        return self._beam_factory.make_polarized_beam(
            sample_to_source=(0.0, 0.0, 1.0),
            wavelength=wavelength,
            polarization=(0, 1, 0),
            polarization_fraction=0.5,
            probe=probe,
        )

    def get_detector(self, index=None):
        return Format.get_detector(self)

    def _detector(self):
        theta_rad = self._gonio_start_angles[1] / 180 * np.pi
        detector_rotns_rad = np.array(self._bin_header["detector_rotns"]) / 180 * np.pi

        # I don't know why only rot_e1 is clockwise but this matches
        # DIRECTION_OF_DETECTOR_X-AXIS/Y-AXIS in XDS.INP from CrysAlisPro.
        # Note that XDS.INP's directions of Y and Z are opposite from ours.
        rot_e1 = np.array(
            r3_rotation_axis_and_angle_as_matrix([0, 0, 1], detector_rotns_rad[0])
        ).reshape(3, 3)  # clockwise along e1 = Z
        rot_e2 = np.array(
            r3_rotation_axis_and_angle_as_matrix([-1, 0, 0], detector_rotns_rad[1])
        ).reshape(3, 3)  # ANTI-clockwise along e2 = X
        rot_theta = np.array(
            r3_rotation_axis_and_angle_as_matrix([0, -1, 0], theta_rad)
        ).reshape(3, 3)  # ANTI-clockwise along e3 = Y
        detector_axes = rot_theta.dot(rot_e2.dot(rot_e1))

        pixel_size_x = self._bin_header["real_px_size_x"]
        pixel_size_y = self._bin_header["real_px_size_y"]
        origin_at_zero = np.array(
            [
                -self._bin_header["origin_px_x"]
                * pixel_size_x
                / self._bin_header["bin_x"],
                -self._bin_header["origin_px_y"]
                * pixel_size_y
                / self._bin_header["bin_y"],
                -self._bin_header["distance_mm"],
            ]
        )
        origin = detector_axes.dot(origin_at_zero)

        detector = self._detector_factory.make_detector(
            "PAD",
            detector_axes[:, 0],
            detector_axes[:, 1],
            origin,
            (pixel_size_x, pixel_size_y),
            (self._txt_header["NX"], self._txt_header["NY"]),
            (0, self._bin_header["overflow_threshold"]),
            gain=self._bin_header["gain"],
        )

        return detector

    def get_scan(self, index=None):
        return Format.get_scan(self)

    def _scan(self):
        """Return the scan information for this image."""

        start_angle = self._gonio_start_angles[self._scan_axis]
        end_angle = self._gonio_end_angles[self._scan_axis]

        return self._scan_factory.single_file(
            filename=self._image_file,
            exposure_times=self._bin_header["exposure_time_sec"],
            osc_start=start_angle,
            osc_width=end_angle - start_angle,
            epoch=None,
        )

    def get_raw_data(self):
        comp = self._txt_header["compression"].strip()

        # The differences between TY5 and TY6 are:
        # - TY5 lacks the lbytesincompressedfield field; we must calculate it.
        # - TY6 uses fewer-bit compression inside "blocks".
        #   Outside such blocks, TY5 and TY6 use the same delta compression.
        if comp.startswith("TY6"):
            return self._get_raw_data_ty5_ty6_native(mode=6)
        elif comp.startswith("TY5"):
            return self._get_raw_data_ty5_ty6_native(mode=5)
        else:
            raise NotImplementedError(f"Can't handle compression: {comp}")

    def _get_raw_data_ty5_ty6_native(self, mode):
        offset = self._txt_header["NHEADER"]
        nx = self._txt_header["NX"]
        ny = self._txt_header["NY"]
        with FormatROD.open_file(self._image_file, "rb") as f:
            f.seek(offset)
            if mode == 5:
                lbytesincompressedfield = (
                    nx * ny + 2 * self._txt_header["OI"] + 4 * self._txt_header["OL"]
                )
            elif mode == 6:
                lbytesincompressedfield = struct.unpack("<l", f.read(4))[0]
            else:
                raise NotImplementedError(
                    "_get_raw_data_ty5_ty6: only mode 5 and 6 are supported."
                )

            linedata = f.read(lbytesincompressedfield)
            offsets = f.read(4 * ny)

            return uncompress_rod_TY5_TY6(linedata, offsets, ny, nx, mode)

    # Python implementation
    def _get_raw_data_ty5_ty6(self, mode):
        offset = self._txt_header["NHEADER"]
        nx = self._txt_header["NX"]
        ny = self._txt_header["NY"]
        with FormatROD.open_file(self._image_file, "rb") as f:
            f.seek(offset)
            if mode == 5:
                lbytesincompressedfield = (
                    nx * ny + 2 * self._txt_header["OI"] + 4 * self._txt_header["OL"]
                )
            elif mode == 6:
                lbytesincompressedfield = struct.unpack("<l", f.read(4))[0]
            else:
                raise NotImplementedError(
                    "_get_raw_data_ty5_ty6: only mode 5 and 6 are supported."
                )

            linedata = np.fromfile(f, dtype=np.uint8, count=lbytesincompressedfield)
            offsets = struct.unpack("<%dI" % ny, f.read(4 * ny))

            image = np.zeros((ny, nx), dtype=np.int32)
            for iy in range(ny):
                image[iy, :] = self.decode_TY5_TY6_oneline(
                    linedata[offsets[iy] :], nx, mode
                )
            return flex.int(image)

    def decode_TY5_TY6_oneline(self, linedata, w, mode):
        """Decompress TY6 encoded pixels for a single line.
        w is the number of pixels in the fast axis."""

        BLOCKSIZE = 8
        SHORT_OVERFLOW = np.int32(254)
        LONG_OVERFLOW = 255
        SHORT_OVERFLOW_SIGNED = SHORT_OVERFLOW - 127
        LONG_OVERFLOW_SIGNED = LONG_OVERFLOW - 127

        ipos = 0
        opos = 0
        ret = np.zeros(w, dtype=np.int32)

        if mode == 5:
            nblock = 0
            nrest = w - 1  # 1st px is special
        elif mode == 6:
            nblock = (w - 1) // (BLOCKSIZE * 2)
            nrest = (w - 1) % (BLOCKSIZE * 2)
        else:
            raise NotImplementedError(
                "decode_TY5_TY6_oneline: only mode 5 and 6 are supported."
            )

        firstpx = linedata[ipos]
        ipos += 1
        if firstpx < SHORT_OVERFLOW:
            ret[opos] = firstpx - 127
        elif firstpx == LONG_OVERFLOW:
            ret[opos] = linedata[ipos : (ipos + 4)].view(np.int32)[0]
            ipos += 4
        else:
            ret[opos] = linedata[ipos : (ipos + 2)].view(np.int16)[0]
            ipos += 2
        opos += 1

        for k in range(nblock):
            bittype = linedata[ipos]
            nbits = (bittype & 15, (bittype >> 4) & 15)
            ipos += 1

            for i in range(2):
                nbit = nbits[i]
                # 0: 0
                # 1: 1 or 0
                # 2: -1, 0, 1, 2
                # 3: -3, -2, -1, 0, 1, 2, 3, 4 (= b111 -3 = 7 - 3)
                zero_at = 0
                if nbit > 1:
                    zero_at = (1 << (nbit - 1)) - 1

                v = 0
                for j in range(nbit):
                    v |= linedata[ipos] << (8 * j)
                    ipos += 1

                mask = (1 << nbit) - 1
                for j in range(BLOCKSIZE):
                    ret[opos] = ((v >> (nbit * j)) & mask) - zero_at
                    opos += 1

            for i in range(opos - BLOCKSIZE * 2, opos):
                offset = ret[i]

                # surprisingly, this IF is very slow!
                if offset >= SHORT_OVERFLOW_SIGNED:
                    if offset >= LONG_OVERFLOW_SIGNED:
                        offset = linedata[ipos : (ipos + 4)].view(np.int32)[0]
                        ipos += 4
                    else:
                        offset = linedata[ipos : (ipos + 2)].view(np.int16)[0]
                        ipos += 2

                ret[i] = offset + ret[i - 1]

        for i in range(nrest):
            px = linedata[ipos]
            ipos += 1
            if px < SHORT_OVERFLOW:
                ret[opos] = ret[opos - 1] + px - 127
            elif px == LONG_OVERFLOW:
                ret[opos] = (
                    ret[opos - 1] + linedata[ipos : (ipos + 4)].view(np.int32)[0]
                )
                ipos += 4
            else:
                ret[opos] = (
                    ret[opos - 1] + linedata[ipos : (ipos + 2)].view(np.int16)[0]
                )
                ipos += 2
            opos += 1

        return ret


class FormatROD_Arc(FormatROD):
    @staticmethod
    def understand(image_file):
        offset = 256
        general_nbytes = 512
        with FormatROD_Arc.open_file(image_file, "rb") as f:
            f.seek(offset + general_nbytes + 548)
            detector_type = struct.unpack("<l", f.read(4))[0]
            if detector_type in [12, 14]:
                return True
        return False

    @staticmethod
    def _read_binary_header(
        image_file,
        offset=256,
        nbytes=5120,
        general_nbytes=512,
        special_nbytes=768,
        km4gonio_nbytes=1024,
        statistics_nbytes=512,
        history_nbytes=2048,
    ):
        bh = FormatROD._read_binary_header(
            image_file,
            offset,
            nbytes,
            general_nbytes,
            special_nbytes,
            km4gonio_nbytes,
            statistics_nbytes,
            history_nbytes,
        )

        # Seek to the end of the standard version 3 header and into the
        # extra camera parameters section
        with FormatROD_Arc.open_file(image_file, "rb") as f:
            f.seek(nbytes + 268)
            ix, iy, nx, ny, gapx, gapy = struct.unpack("<hhhhhh", f.read(12))
        assert ix in (2, 3)
        assert iy == 1
        assert nx == 385
        assert ny == 775
        assert gapx == 30
        assert gapy == 0
        bh["nx"] = nx
        bh["ny"] = ny
        bh["gap_px"] = gapx

        return bh

    def _detector(self):
        """2 or 3 panel detector, each rotated 38° from its neighbour."""

        theta_rad = self._gonio_start_angles[1] / 180 * np.pi
        detector_rotns_rad = np.array(self._bin_header["detector_rotns"]) / 180 * np.pi
        rot_e1 = np.array(
            r3_rotation_axis_and_angle_as_matrix([0, 0, 1], detector_rotns_rad[0])
        ).reshape(3, 3)  # clockwise along e1 = Z
        rot_e2 = np.array(
            r3_rotation_axis_and_angle_as_matrix([-1, 0, 0], detector_rotns_rad[1])
        ).reshape(3, 3)  # ANTI-clockwise along e2 = X
        rot_theta = np.array(
            r3_rotation_axis_and_angle_as_matrix([0, -1, 0], theta_rad)
        ).reshape(3, 3)  # ANTI-clockwise along e3 = Y
        detector_axes = rot_theta.dot(rot_e2.dot(rot_e1))

        pixel_size_x = self._bin_header["real_px_size_x"]
        pixel_size_y = self._bin_header["real_px_size_y"]
        origin_at_zero = np.array(
            [
                -self._bin_header["origin_px_x"] * pixel_size_x,
                -self._bin_header["origin_px_y"] * pixel_size_y,
                -self._bin_header["distance_mm"],
            ]
        )
        # Get origin for the first panel, not rotated
        origin = detector_axes.dot(origin_at_zero)

        d = Detector()

        root = d.hierarchy()
        root.set_local_frame(detector_axes[:, 0], detector_axes[:, 1], origin)

        self.coords = {}
        panel_idx = 0

        pnl_data = []
        nx = self._bin_header["nx"]
        ny = self._bin_header["ny"]
        gap_px = self._bin_header["gap_px"]
        pnl_data.append({"xmin": 0, "ymin": 0, "xmax": nx, "ymax": ny, "xmin_mm": 0})
        pnl_data.append(
            {
                "xmin": (nx + gap_px),
                "ymin": 0,
                "xmax": 2 * nx + gap_px,
                "ymax": ny,
                "xmin_mm": (nx + gap_px) * pixel_size_x,
            }
        )
        if self._bin_header["detector_type"] == 12:
            pnl_data.append(
                {
                    "xmin": (nx + gap_px) * 2,
                    "ymin": 0,
                    "xmax": 3 * nx + 2 * gap_px,
                    "ymax": ny,
                    "xmin_mm": (nx + gap_px) * 2 * pixel_size_x,
                }
            )

        # redefine fast, slow for the local frame
        fast = matrix.col((1.0, 0.0, 0.0))
        slow = matrix.col((0.0, 1.0, 0.0))

        for ipanel, pd in enumerate(pnl_data):
            xmin = pd["xmin"]
            xmax = pd["xmax"]
            ymin = pd["ymin"]
            ymax = pd["ymax"]
            xmin_mm = pd["xmin_mm"]

            origin_panel = fast * xmin_mm

            panel_name = "Panel%d" % panel_idx
            panel_idx += 1

            p = d.add_panel()
            p.set_type("SENSOR_PAD")
            p.set_name(panel_name)
            p.set_gain(self._bin_header["gain"])
            p.set_raw_image_offset((xmin, ymin))
            p.set_image_size((xmax - xmin, ymax - ymin))
            p.set_trusted_range((0, self._bin_header["overflow_threshold"]))
            p.set_pixel_size((pixel_size_x, pixel_size_y))
            p.set_local_frame(fast.elems, slow.elems, origin_panel.elems)
            p.set_projection_2d((-1, 0, 0, -1), (0, pd["xmin"]))
            self.coords[panel_name] = (xmin, ymin, xmax, ymax)

        # Now rotate the external panels. For the time being do the rotation around
        # and axis along the centre of the gap between the panels.
        if len(pnl_data) == 2:
            angle = 19.0
        elif len(pnl_data) == 3:
            angle = 38.0
        else:
            raise ValueError("Unexpected number of panels")

        left_panel = d[0]
        right_panel = d[len(d) - 1]

        # Point to rotate the left panel around, from the local origin
        pnl_fast = matrix.col(left_panel.get_local_fast_axis())
        pnl_slow = matrix.col(left_panel.get_local_slow_axis())
        pt = pnl_fast * (
            left_panel.get_image_size_mm()[0] + (gap_px / 2) * pixel_size_x
        )

        rotated = (-1.0 * pt).rotate_around_origin(pnl_slow, angle, deg=True)
        new_origin = pt + rotated
        new_fast = pnl_fast.rotate_around_origin(pnl_slow, angle, deg=True)
        left_panel.set_local_frame(new_fast.elems, pnl_slow.elems, new_origin.elems)

        # Point to rotate the right panel around, from the panel's origin
        pnl_fast = matrix.col(right_panel.get_local_fast_axis())
        pnl_slow = matrix.col(right_panel.get_local_slow_axis())
        pnl_origin = matrix.col(right_panel.get_local_origin())
        pt = -pnl_fast * (gap_px / 2) * pixel_size_x

        rotated = (-1.0 * pt).rotate_around_origin(pnl_slow, -angle, deg=True)
        new_origin = pnl_origin + pt + rotated
        new_fast = pnl_fast.rotate_around_origin(pnl_slow, -angle, deg=True)
        right_panel.set_local_frame(new_fast.elems, pnl_slow.elems, new_origin.elems)

        return d

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        raw_data = super().get_raw_data()

        # split into separate panels
        self._raw_data = []
        d = self.get_detector()
        for panel in d:
            xmin, ymin, xmax, ymax = self.coords[panel.get_name()]
            self._raw_data.append(raw_data[ymin:ymax, xmin:xmax])

        return tuple(self._raw_data)


if __name__ == "__main__":
    import sys

    for arg in sys.argv[1:]:
        if FormatROD.understand(arg):
            reader = FormatROD(arg)
            print(reader._txt_header)
            print(reader._bin_header)
            print()
            print(
                "Starting angles in degrees (OMEGA, THETA, KAPPA=CHI, PHI, OMEGA PRIME, THETA PRIME)"
            )
            print(reader._gonio_start_angles)
            print("Ending angles in degrees")
            print(reader._gonio_end_angles)
        else:
            print("Unsupported format")



from __future__ import annotations

import pytest

from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.model import (
    Beam,
    BeamFactory,
    Goniometer,
    GoniometerFactory,
    Scan,
    ScanFactory,
)


def test_beam():
    b1 = Beam((1, 0, 0), 2, 0.1, 0.1)
    d = b1.to_dict()
    b2 = BeamFactory.from_dict(d)
    assert d["direction"] == (1, 0, 0)
    assert d["wavelength"] == 2
    assert d["divergence"] == pytest.approx(0.1)
    assert d["sigma_divergence"] == pytest.approx(0.1)
    assert b1 == b2
    assert "s0_at_scan_points" not in d

    # Test with a template and partial dictionary
    d2 = {"direction": (0, 1, 0), "divergence": 0.2}
    b3 = BeamFactory.from_dict(d2, d)
    assert b3.get_sample_to_source_direction() == (0, 1, 0)
    assert b3.get_wavelength() == 2
    assert b3.get_divergence() == pytest.approx(0.2)
    assert b3.get_sigma_divergence() == pytest.approx(0.1)
    assert b2 != b3


def test_beam_with_scan_points():
    b1 = Beam((1, 0, 0), 2, 0.1, 0.1)

    s0_static = matrix.col(b1.get_s0())
    b1.set_s0_at_scan_points([s0_static] * 5)
    d = b1.to_dict()
    b2 = BeamFactory.from_dict(d)

    for s0comp in d["s0_at_scan_points"]:
        assert matrix.col(s0comp) == s0_static

    for s0comp in b2.get_s0_at_scan_points():
        assert matrix.col(s0comp) == s0_static

    assert b1 == b2


def test_goniometer():
    g1 = Goniometer((1, 0, 0), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    d = g1.to_dict()
    g2 = GoniometerFactory.from_dict(d)
    assert d["rotation_axis"] == (1, 0, 0)
    assert d["fixed_rotation"] == (1, 0, 0, 0, 1, 0, 0, 0, 1)
    assert g1 == g2
    assert "setting_rotation_at_scan_points" not in d

    # Test with a template and partial dictionary
    d2 = {"rotation_axis": (0, 1, 0)}
    g3 = GoniometerFactory.from_dict(d2, d)
    assert g3.get_rotation_axis() == (0, 1, 0)
    assert g3.get_fixed_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)
    assert g2 != g3


def test_multi_axis_goniometer():
    g1 = GoniometerFactory.multi_axis(
        flex.vec3_double(((1, 0, 0),)), flex.double((0,)), flex.std_string(("PHI",)), 0
    )
    d = g1.to_dict()
    g2 = GoniometerFactory.from_dict(d)
    assert d["axes"] == [(1, 0, 0)]
    assert d["angles"] == [0.0]
    assert d["names"] == ["PHI"]
    assert d["scan_axis"] == 0
    assert g1 == g2
    assert "setting_rotation_at_scan_points" not in d

    # Test with a template and partial dictionary
    d2 = {"axes": [(0, 1, 0)]}
    g3 = GoniometerFactory.from_dict(d2, d)
    assert g3.get_rotation_axis() == (0, 1, 0)
    assert g3.get_fixed_rotation() == (1, 0, 0, 0, 1, 0, 0, 0, 1)
    assert g2 != g3


def test_goniometer_with_scan_points():
    simple_g = Goniometer((1, 0, 0), (1, 0, 0, 0, 1, 0, 0, 0, 1))
    multi_ax_g = GoniometerFactory.multi_axis(
        flex.vec3_double(((1, 0, 0),)), flex.double((0,)), flex.std_string(("PHI",)), 0
    )

    for g1 in [simple_g, multi_ax_g]:
        S_static = matrix.sqr(g1.get_setting_rotation())
        g1.set_setting_rotation_at_scan_points([S_static] * 5)
        d = g1.to_dict()
        g2 = GoniometerFactory.from_dict(d)

        for Scomp in d["setting_rotation_at_scan_points"]:
            assert matrix.sqr(Scomp) == S_static

        for Scomp in g2.get_setting_rotation_at_scan_points():
            assert matrix.sqr(Scomp) == S_static

        assert g1 == g2


def test_scan():
    s1 = Scan(
        (1, 3),
        (1.0, 0.2),
        flex.double([0.1, 0.1, 0.1]),
        flex.double([0.1, 0.2, 0.3]),
        0,
    )
    d = s1.to_dict()
    s2 = ScanFactory.from_dict(d)
    assert d["image_range"] == (1, 3)
    osc = d["properties"]["oscillation"]
    assert (osc[0], osc[1] - osc[0]) == pytest.approx((1.0, 0.2))
    assert d["properties"]["exposure_time"] == pytest.approx([0.1, 0.1, 0.1])
    assert d["properties"]["epochs"] == pytest.approx([0.1, 0.2, 0.3])
    assert d["batch_offset"] == 0
    assert s1 == s2

    # Test with a template and partial dictionary
    d2 = {"exposure_time": [0.2, 0.2, 0.2]}
    s3 = ScanFactory.from_dict(d2, d)
    assert s3.get_image_range() == (1, 3)
    assert s3.get_oscillation() == pytest.approx((1.0, 0.2))
    assert list(s3.get_exposure_times()) == pytest.approx([0.2, 0.2, 0.2])
    assert list(s3.get_epochs()) == pytest.approx([0.1, 0.2, 0.3])
    assert s2 != s3

    # Test with a partial epoch
    d3 = {"image_range": (1, 10), "epochs": [0.1, 0.2]}
    s4 = ScanFactory.from_dict(d3, d)
    assert s4.get_epochs()[2] == pytest.approx(0.3)
    assert s4.get_epochs()[9] == pytest.approx(1.0)

    d4 = {"batch_offset": 100}
    s5 = ScanFactory.from_dict(d4, d)
    assert s5.get_batch_offset() == 100
    assert s5.get_batch_range() == (101, 103)



from __future__ import annotations

import calendar

import h5py

from dxtbx.format.FormatCBFMiniPilatusHelpers import get_pilatus_timestamp
from dxtbx.format.FormatNXmxDLS16M import FormatNXmxDLS16M
from dxtbx.format.nexus import h5str


class FormatNXmxDLS16MI03(FormatNXmxDLS16M):
    """Format class for the Eiger 2XE 16M on DLS I03

    Only active for data collected between 20210422 and 20210503 (inclusive).
    Mask out two known bad modules in full detector mode, and one module in
    ROI mode.
    """

    @staticmethod
    def understand(image_file):
        # Get the file handle
        with h5py.File(image_file) as fh:
            name = FormatNXmxDLS16M.get_instrument_name(fh)
            if name and name.upper() in {"DIAMOND BEAMLINE I03", "DLS I03"}:
                timestamp = get_pilatus_timestamp(
                    h5str(fh["/entry/start_time"][()]).rstrip("Z")
                )
                return timestamp > calendar.timegm(
                    (2021, 4, 22, 0, 0, 0)
                ) and timestamp <= calendar.timegm((2021, 6, 22, 0, 0, 0))

    def _start(self):
        super()._start()
        nx = 1028  # module pixels x
        ny = 512  # module pixels y
        dx = 12  # module gap size
        dy = 38  # module gap size
        if self._detector_model[0].get_image_size() == (4148, 4362):
            # Full detector mode
            bad_modules = {(2, 5), (3, 5)}
        else:
            # ROI detector mode
            bad_modules = {(1, 3)}
        for i_col, i_row in bad_modules:
            self._detector_model[0].add_mask(
                (nx + dx) * i_col,
                (ny + dy) * i_row,
                (nx + dx) * i_col + nx,
                (ny + dy) * i_row + ny,
            )



"""Implementation of a base class to read a pickled Python dictionary."""

from __future__ import annotations

import pickle
import sys

from dxtbx import IncorrectFormatError
from dxtbx.format.Format import Format, abstract


@abstract
class FormatPY(Format):
    """Let's take an educated guess as to how to recognize a Python
    pickle file containing a dictionary.  Not easy because there are
    six pickle protocols in Python.  The lowest pickle format only
    gives us two unique bytes by which to recognize a dictionary,
    so also check if we can unpickle the file."""

    @staticmethod
    def understand(image_file):
        try:
            with FormatPY.open_file(image_file, "rb") as fh:
                tag = fh.read(4)
                if (  # pickle protocols 0-5 where the pickled object is a dictionary
                    tag[0:2] == b"(d"
                    or tag[0:2] == b"}q"
                    or tag[0:4] == b"\x80\x02}q"
                    or tag[0:4] == b"\x80\x03}q"
                    or tag[0:3] == b"\x80\x04\x95"
                    or tag[0:3] == b"\x80\x05\x95"
                ):
                    fh.seek(0)
                    pickle.load(fh, encoding="bytes")
                else:
                    return False
        except (OSError, pickle.UnpicklingError):
            return False
        return True

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        if not self.understand(image_file):
            raise IncorrectFormatError(self, image_file)
        super().__init__(image_file, **kwargs)


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatPY.understand(arg))



# LIBTBX_SET_DISPATCHER_NAME dxtbx.image_average
# LIBTBX_SET_DISPATCHER_NAME cxi.image_average

"""
Average images of any dxtbx-supported format. Handles many individual images or single container files.
"""

from __future__ import annotations

import copy
import os
import sys

import numpy as np

from libtbx import easy_mp, option_parser
from libtbx.utils import Sorry, Usage
from scitbx.array_family import flex

import dxtbx.format.Registry
import dxtbx.util
from dxtbx.format.cbf_writer import FullCBFWriter
from dxtbx.format.FormatMultiImage import FormatMultiImage
from dxtbx.model.experiment_list import ExperimentListFactory


def splitit(l, n):
    """Utility function to evenly split a list. Handles edge cases.
    There is probably a 1-liner list comprehension to do this, but it would be super gnarly.
    @param l list to split (not a generator)
    @param n number of chunks to split the list into
    @return list of n lists
    """
    # if list is shorter then n, split into a list of lists, each with one entry
    if len(l) < n:
        n = len(l)
    s = len(l) // (n)  # each chunk will either be of size s or size s+1
    m = len(l) % (
        s * n
    )  # remainder after n chunks of size s = how many chunks of size s+1 are needed
    r = []  # result
    p = 0  # pointer
    for i in range(n):
        if i < m:
            r.append(l[p : p + s + 1])
            p += s + 1
        else:
            r.append(l[p : p + s])
            p += s
    return r


class image_worker:
    """Class to compute running sums while reading image data"""

    # Deriving class should implement __init__ and read

    def __call__(self, subset):
        """Worker function for multiprocessing"""
        nfail = 0
        nmemb = 0

        for item in subset:
            try:
                img, distance, wavelength = self.read(item)
            except Exception as e:
                print(str(e))
                nfail += 1
                continue

            assert isinstance(img, tuple)

            # The sum-of-squares image is accumulated using long integers, as
            # this delays the point where overflow occurs.  But really, this
            # is just a band-aid...
            if nmemb == 0:
                max_img = list(copy.deepcopy(img))
                sum_distance = distance
                sum_img = list(copy.deepcopy(img))
                ssq_img = [flex.pow2(p) for p in img]
                sum_wavelength = wavelength
            else:
                for n, image in enumerate(img):
                    sel = (image > max_img[n]).as_1d()
                    max_img[n].set_selected(sel, image.select(sel))

                    sum_img[n] += image
                    ssq_img[n] += flex.pow2(image)
                sum_distance += distance
                sum_wavelength += wavelength
            nmemb += 1

        return nfail, nmemb, max_img, sum_distance, sum_img, ssq_img, sum_wavelength


class multi_image_worker(image_worker):
    """Class for reading container files"""

    def __init__(self, command_line, path, experiments):
        self.path = path
        self.command_line = command_line

        self.experiments = experiments

    def read(self, n):
        """Read image at position n"""
        if self.command_line.options.verbose:
            print("Processing %s: %d" % (self.path, n))

        expt = self.experiments[n]
        expt.load_models()

        beam = expt.beam
        detector = expt.detector

        image_data = expt.imageset[0]
        if not isinstance(image_data, tuple):
            image_data = (image_data,)
        img = tuple(image_data[i].as_1d().as_double() for i in range(len(detector)))
        wavelength = beam.get_wavelength()
        expt.imageset.clear_cache()

        return img, detector.hierarchy().get_distance(), wavelength


class single_image_worker(image_worker):
    """Class for averaging single images from individual files"""

    def __init__(self, command_line):
        self.command_line = command_line

    def read(self, path):
        if self.command_line.options.verbose:
            print("Processing %s" % path)

        format_class = dxtbx.format.Registry.get_format_class_for_file(path)
        assert not issubclass(format_class, FormatMultiImage), (
            "Average container files separately"
        )
        img_instance = format_class(path)

        beam = img_instance.get_beam()
        detector = img_instance.get_detector()
        image_data = img_instance.get_raw_data()
        if not isinstance(image_data, tuple):
            image_data = (image_data,)
        img = tuple(image_data[i].as_1d().as_double() for i in range(len(detector)))
        wavelength = beam.get_wavelength()

        return img, detector.hierarchy().get_distance(), wavelength


def run(argv=None):
    """Compute mean, standard deviation, and maximum projection images
    from a set of images given on the command line.

    @param argv Command line argument list
    @return     @c 0 on successful termination, @c 1 on error, and @c 2
                for command line syntax errors
    """
    if argv is None:
        argv = sys.argv
    dxtbx.util.encode_output_as_utf8()
    progname = os.getenv("LIBTBX_DISPATCHER_NAME")
    if not progname or progname.endswith(".python"):
        progname = "%prog"
    command_line = (
        option_parser.option_parser(
            usage=f"{progname} [-v] [-a PATH] [-m PATH] [-s PATH] image1 image2 [image3 ...]"
        )
        .option(
            None,
            "--average-path",
            "-a",
            type="string",
            default="avg.cbf",
            dest="avg_path",
            metavar="PATH",
            help="Write average image to PATH",
        )
        .option(
            None,
            "--maximum-path",
            "-m",
            type="string",
            default="max.cbf",
            dest="max_path",
            metavar="PATH",
            help="Write maximum projection image to PATH",
        )
        .option(
            None,
            "--stddev-path",
            "-s",
            type="string",
            default="stddev.cbf",
            dest="stddev_path",
            metavar="PATH",
            help="Write standard deviation image to PATH",
        )
        .option(
            None,
            "--verbose",
            "-v",
            action="store_true",
            default=False,
            dest="verbose",
            help="Print more information about progress",
        )
        .option(
            None,
            "--nproc",
            "-n",
            type="int",
            default=1,
            dest="nproc",
            help="Number of processors",
        )
        .option(
            None,
            "--num-images-max",
            "-N",
            type="int",
            default=None,
            dest="num_images_max",
            help="Maximum number of frames to average",
        )
        .option(
            None,
            "--skip-images",
            "-S",
            type="int",
            default=None,
            dest="skip_images",
            help="Number of images to skip at the start of the dataset",
        )
        .option(
            None,
            "--mpi",
            None,
            type=bool,
            default=False,
            dest="mpi",
            metavar="False",
            help="Set to True enable MPI processing",
        )
    ).process(args=argv[1:])

    # Note that it is not an error to omit the output paths, because
    # certain statistics could still be printed, e.g. with the verbose
    # option.
    paths = command_line.args
    if len(paths) == 0:
        command_line.parser.print_usage(file=sys.stderr)
        return 2

    if any(os.path.splitext(p)[1].lower() == ".loc" for p in paths):
        import psana

        if getattr(psana, "xtc_version", None) == 2:
            # multiprocessing with psana2 requires mpi
            assert command_line.options.mpi or command_line.options.nproc == 1
            root = 2  # psana2 uses ranks 0 and 1
        else:
            root = 0
    else:
        root = 0

    experiments = ExperimentListFactory.from_filenames([paths[0]], load_models=False)
    if len(paths) == 1:
        if len(experiments) == 1 and len(experiments[0].imageset) > 1:
            experiments = ExperimentListFactory.from_stills_and_crystal(
                experiments[0].imageset, None, load_models=False
            )
        worker = multi_image_worker(command_line, paths[0], experiments)
        iterable = list(range(len(experiments)))
    else:
        # Multiple images provided
        worker = single_image_worker(command_line)
        iterable = paths

    if command_line.options.skip_images is not None:
        if command_line.options.skip_images >= len(iterable):
            raise Usage("Skipping all the images")
        iterable = iterable[command_line.options.skip_images :]
    if (
        command_line.options.num_images_max is not None
        and command_line.options.num_images_max < len(iterable)
    ):
        iterable = iterable[: command_line.options.num_images_max]

    if command_line.options.mpi:
        try:
            from libtbx.mpi4py import MPI
        except ImportError:
            raise Sorry("MPI not found")
        comm = MPI.COMM_WORLD
        rank = comm.Get_rank()
        size = comm.Get_size() - root  # handle psana2

        # chop the list into pieces, depending on rank.  This assigns each process
        # events such that the get every Nth event where N is the number of processes
        iterable_rank = iterable[rank - root :: size]
        if len(iterable_rank) > 0:
            # Only run the worker on non-empty ranks
            (
                r_nfail,
                r_nmemb,
                r_max_img,
                r_sum_distance,
                r_sum_img,
                r_ssq_img,
                r_sum_wavelength,
            ) = worker(iterable_rank)

            # send the shape of the data to those ranks with no data
            shapes_ = [p.focus() for p in r_sum_img]
            shapes = comm.allgather(shapes_)
        else:
            # else set the values to zero
            r_nfail = 0
            r_nmemb = 0
            r_sum_distance = 0.0
            r_sum_wavelength = 0.0
            shapes = comm.allgather(None)
            shape = [ss for ss in shapes if ss is not None][0]
            r_sum_img = [flex.double(flex.grid(s), 0) for s in shape]
            r_ssq_img = [flex.double(flex.grid(s), 0) for s in shape]
            r_max_img = [flex.double(flex.grid(s), 0) for s in shape]

        nfail = np.array([0])
        nmemb = np.array([0])
        sum_distance = np.array([0.0])
        sum_wavelength = np.array([0.0])
        comm.Reduce(np.array([r_nfail]), nfail, root=root)
        comm.Reduce(np.array([r_nmemb]), nmemb, root=root)
        comm.Reduce(np.array([r_sum_distance]), sum_distance, root=root)
        comm.Reduce(np.array([r_sum_wavelength]), sum_wavelength, root=root)
        nfail = int(nfail[0])
        nmemb = int(nmemb[0])
        sum_distance = float(sum_distance[0])
        sum_wavelength = float(sum_wavelength[0])

        def reduce_image(data, op=MPI.SUM):
            result = []
            for panel_data in data:
                panel_data = panel_data.as_numpy_array()
                reduced_data = np.zeros(panel_data.shape).astype(panel_data.dtype)
                comm.Reduce(panel_data, reduced_data, op=op, root=root)
                result.append(flex.double(reduced_data))
            return result

        max_img = reduce_image(r_max_img, MPI.MAX)
        sum_img = reduce_image(r_sum_img)
        ssq_img = reduce_image(r_ssq_img)

        if rank != root:
            return
        avg_img = tuple(s / nmemb for s in sum_img)
    else:
        if command_line.options.nproc == 1:
            results = [worker(iterable)]
        else:
            iterable = splitit(iterable, command_line.options.nproc)
            results = easy_mp.parallel_map(
                func=worker, iterable=iterable, processes=command_line.options.nproc
            )

        nfail = 0
        nmemb = 0
        for (
            i,
            (
                r_nfail,
                r_nmemb,
                r_max_img,
                r_sum_distance,
                r_sum_img,
                r_ssq_img,
                r_sum_wavelength,
            ),
        ) in enumerate(results):
            nfail += r_nfail
            nmemb += r_nmemb
            if i == 0:
                max_img = r_max_img
                sum_distance = r_sum_distance
                sum_img = r_sum_img
                ssq_img = r_ssq_img
                sum_wavelength = r_sum_wavelength
            else:
                for p in range(len(sum_img)):
                    sel = (r_max_img[p] > max_img[p]).as_1d()
                    max_img[p].set_selected(sel, r_max_img[p].select(sel))

                    sum_img[p] += r_sum_img[p]
                    ssq_img[p] += r_ssq_img[p]

                sum_distance += r_sum_distance
                sum_wavelength += r_sum_wavelength

    # Early exit if no statistics were accumulated.
    if command_line.options.verbose:
        sys.stdout.write("Processed %d images (%d failed)\n" % (nmemb, nfail))
    if nmemb == 0:
        return 0

    # Calculate averages for measures where other statistics do not make
    # sense.  Note that avg_img is required for stddev_img.
    avg_img = tuple(s.as_double() / nmemb for s in sum_img)
    avg_distance = sum_distance / nmemb
    avg_wavelength = sum_wavelength / nmemb
    expt = experiments[0]
    expt.load_models()
    detector = expt.detector
    h = detector.hierarchy()
    origin = h.get_local_origin()
    h.set_local_frame(
        h.get_local_fast_axis(),
        h.get_local_slow_axis(),
        (origin[0], origin[1], -avg_distance),
    )
    expt.beam.set_wavelength(avg_wavelength)
    assert expt.beam.get_wavelength() == expt.imageset.get_beam(0).get_wavelength()

    # Output the average image, maximum projection image, and standard
    # deviation image, if requested.
    if command_line.options.avg_path is not None:
        for n, d in enumerate(detector):
            fast, slow = d.get_image_size()
            avg_img[n].resize(flex.grid(slow, fast))
        writer = FullCBFWriter(imageset=expt.imageset)
        cbf = writer.get_cbf_handle(header_only=True)
        writer.add_data_to_cbf(cbf, data=avg_img)
        writer.write_cbf(command_line.options.avg_path, cbf=cbf)

    if command_line.options.max_path is not None:
        for n, d in enumerate(detector):
            fast, slow = d.get_image_size()
            max_img[n].resize(flex.grid(slow, fast))
        max_img = tuple(max_img)

        writer = FullCBFWriter(imageset=expt.imageset)
        cbf = writer.get_cbf_handle(header_only=True)
        writer.add_data_to_cbf(cbf, data=max_img)
        writer.write_cbf(command_line.options.max_path, cbf=cbf)

    if command_line.options.stddev_path is not None:
        stddev_img = []
        for n, d in enumerate(detector):
            stddev_img.append(
                ssq_img[n].as_double() - sum_img[n].as_double() * avg_img[n]
            )
            # Accumulating floating-point numbers introduces errors, which may
            # cause negative variances.  Since a two-pass approach is
            # unacceptable, the standard deviation is clamped at zero.
            stddev_img[n].set_selected(stddev_img[n] < 0, 0)
            if nmemb == 1:
                stddev_img[n] = flex.sqrt(stddev_img[n])
            else:
                stddev_img[n] = flex.sqrt(stddev_img[n] / (nmemb - 1))

            fast, slow = d.get_image_size()
            stddev_img[n].resize(flex.grid(slow, fast))
        stddev_img = tuple(stddev_img)

        writer = FullCBFWriter(imageset=expt.imageset)
        cbf = writer.get_cbf_handle(header_only=True)
        writer.add_data_to_cbf(cbf, data=stddev_img)
        writer.write_cbf(command_line.options.stddev_path, cbf=cbf)

    return 0


if __name__ == "__main__":
    sys.exit(run())



"""An implementation of the CBF image reader for Pilatus images"""

from __future__ import annotations

import os
import sys

from cctbx.eltbx import attenuation_coefficient
from scitbx import matrix

from dxtbx.format.FormatCBFMini import FormatCBFMini
from dxtbx.format.FormatCBFMiniPilatusHelpers import get_pilatus_timestamp
from dxtbx.format.FormatPilatusHelpers import _DetectorDatabase, determine_pilatus_mask
from dxtbx.format.FormatPilatusHelpers import get_vendortype as gv
from dxtbx.model import Detector, ParallaxCorrectedPxMmStrategy


class FormatCBFMiniPilatus(FormatCBFMini):
    """A class for reading mini CBF format Pilatus images, and correctly
    constructing a model for the experiment from this."""

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file, including a
        proper model of the experiment."""

        self._multi_panel = kwargs.get("multi_panel", False)
        super().__init__(image_file, **kwargs)

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Pilatus mini CBF format image,
        i.e. we can make sense of it."""

        if "ENABLE_PHOTON_FACTORY_TWO_EIGER" in os.environ:
            return False

        header = FormatCBFMini.get_cbf_header(image_file)

        for record in header.split("\n"):
            if "# Detector" in record:
                if "EIGER" in record.upper():
                    return False
                if "TIMEPIX" in record.upper():
                    return False

        for record in header.split("\n"):
            if "_array_data.header_convention" in record and "PILATUS" in record:
                return True
            if "_array_data.header_convention" in record and "SLS" in record:
                return True
            if "_array_data.header_convention" in record and "?" in record:
                return True
            if "# Detector" in record and "PILATUS" in record:  # CBFlib v0.8.0 allowed
                return True

        return False

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        if not self._multi_panel:
            detector = FormatCBFMini._detector(self)
            for f0, f1, s0, s1 in determine_pilatus_mask(detector):
                detector[0].add_mask(f0 - 1, s0 - 1, f1, s1)
            return detector

        # got to here means 60-panel version
        d = Detector()

        distance = float(self._cif_header_dictionary["Detector_distance"].split()[0])

        beam_xy = (
            self._cif_header_dictionary["Beam_xy"]
            .replace("(", "")
            .replace(")", "")
            .replace(",", "")
            .split()[:2]
        )

        beam_x, beam_y = map(float, beam_xy)

        wavelength = float(self._cif_header_dictionary["Wavelength"].split()[0])

        pixel_xy = (
            self._cif_header_dictionary["Pixel_size"]
            .replace("m", "")
            .replace("x", "")
            .split()
        )

        pixel_x, pixel_y = map(float, pixel_xy)

        thickness = float(self._cif_header_dictionary["Silicon"].split()[2]) * 1000.0

        nx = int(self._cif_header_dictionary["X-Binary-Size-Fastest-Dimension"])
        ny = int(self._cif_header_dictionary["X-Binary-Size-Second-Dimension"])

        max_trusted_value = int(self._cif_header_dictionary["Count_cutoff"].split()[0])
        min_trusted_value = 0

        # take into consideration here the thickness of the sensor also the
        # wavelength of the radiation (which we have in the same file...)
        table = attenuation_coefficient.get_table("Si")
        mu = table.mu_at_angstrom(wavelength) / 10.0
        t0 = thickness

        # FIXME would also be very nice to be able to take into account the
        # misalignment of the individual modules given the calibration...

        # single detector or multi-module detector

        pixel_x *= 1000.0
        pixel_y *= 1000.0
        distance *= 1000.0

        beam_centre = matrix.col((beam_x * pixel_x, beam_y * pixel_y, 0))

        fast = matrix.col((1.0, 0.0, 0.0))
        slow = matrix.col((0.0, -1.0, 0.0))
        s0 = matrix.col((0, 0, -1))
        origin = (distance * s0) - (fast * beam_centre[0]) - (slow * beam_centre[1])

        root = d.hierarchy()
        root.set_local_frame(fast.elems, slow.elems, origin.elems)

        det = _DetectorDatabase["Pilatus"]

        # Edge dead areas not included, only gaps between modules matter
        n_fast, remainder = divmod(nx, det.module_size_fast)
        assert (n_fast - 1) * det.gap_fast == remainder

        n_slow, remainder = divmod(ny, det.module_size_slow)
        assert (n_slow - 1) * det.gap_slow == remainder

        mx = det.module_size_fast
        my = det.module_size_slow
        dx = det.gap_fast
        dy = det.gap_slow

        xmins = [(mx + dx) * i for i in range(n_fast)]
        xmaxes = [mx + (mx + dx) * i for i in range(n_fast)]
        ymins = [(my + dy) * i for i in range(n_slow)]
        ymaxes = [my + (my + dy) * i for i in range(n_slow)]

        self.coords = {}

        fast = matrix.col((1.0, 0.0, 0.0))
        slow = matrix.col((0.0, 1.0, 0.0))
        panel_idx = 0
        for ymin, ymax in zip(ymins, ymaxes):
            for xmin, xmax in zip(xmins, xmaxes):
                xmin_mm = xmin * pixel_x
                ymin_mm = ymin * pixel_y

                origin_panel = fast * xmin_mm + slow * ymin_mm

                panel_name = "Panel%d" % panel_idx
                panel_idx += 1

                p = d.add_panel()
                p.set_type("SENSOR_PAD")
                p.set_name(panel_name)
                p.set_raw_image_offset((xmin, ymin))
                p.set_image_size((xmax - xmin, ymax - ymin))
                p.set_trusted_range((min_trusted_value, max_trusted_value))
                p.set_pixel_size((pixel_x, pixel_y))
                p.set_thickness(thickness)
                p.set_material("Si")
                p.set_mu(mu)
                p.set_px_mm_strategy(ParallaxCorrectedPxMmStrategy(mu, t0))
                p.set_local_frame(fast.elems, slow.elems, origin_panel.elems)
                p.set_raw_image_offset((xmin, ymin))
                self.coords[panel_name] = (xmin, ymin, xmax, ymax)

        return d

    def _scan(self):
        """Return the scan information for this image."""

        exposure_time = float(self._cif_header_dictionary["Exposure_period"].split()[0])

        osc_start = float(self._cif_header_dictionary["Start_angle"].split()[0])
        osc_range = float(self._cif_header_dictionary["Angle_increment"].split()[0])

        timestamp = get_pilatus_timestamp(self._cif_header_dictionary["timestamp"])

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, timestamp
        )

    def get_vendortype(self):
        return gv(self.get_detector())

    @staticmethod
    def as_file(
        detector,
        beam,
        gonio,
        scan,
        data,
        path,
        header_convention="PILATUS_1.2",
        det_type="PILATUS3 6M",
    ):
        FormatCBFMini.as_file(
            detector, beam, gonio, scan, data, path, header_convention, det_type
        )

    def get_raw_data(self):
        if not self._multi_panel:
            return super().get_raw_data()

        if self._raw_data is None:
            raw_data = self._read_cbf_image()
            self._raw_data = []
            d = self.get_detector()
            for panel in d:
                xmin, ymin, xmax, ymax = self.coords[panel.get_name()]
                self._raw_data.append(raw_data[ymin:ymax, xmin:xmax])
            self._raw_data = tuple(self._raw_data)
        return self._raw_data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniPilatus.understand(arg))



from __future__ import annotations

import math
import sys

import libtbx

from dxtbx.format.FormatCBFFullPilatus import FormatCBFFullPilatus
from dxtbx.masking import GoniometerMaskerFactory


class FormatCBFFullPilatusDLS300KSN104(FormatCBFFullPilatus):
    """An image reading class for full CBF format images from Pilatus
    detectors. For DLS I19-2"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        # this depends on DIALS for the goniometer shadow model; if missing
        # simply return False

        header = FormatCBFFullPilatus.get_cbf_header(image_file)

        for record in header.split("\n"):
            if (
                "# Detector" in record
                and "PILATUS 300K" in record
                and "S/N 3-0104, Diamond" in record
            ):
                return True

        return False

    @staticmethod
    def has_dynamic_shadowing(**kwargs):
        dynamic_shadowing = kwargs.get("dynamic_shadowing", False)
        if dynamic_shadowing in (libtbx.Auto, "Auto"):
            return False
        return dynamic_shadowing

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""
        self._dynamic_shadowing = self.has_dynamic_shadowing(**kwargs)
        super().__init__(image_file, **kwargs)

    def get_goniometer_shadow_masker(self, goniometer=None):
        if goniometer is None:
            goniometer = self.get_goniometer()

        return GoniometerMaskerFactory.diamond_anvil_cell(
            goniometer, cone_opening_angle=2 * 38 * math.pi / 180
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFFullPilatusDLS300KSN104.understand(arg))



from __future__ import annotations

import sys

import libtbx

from dxtbx.format.FormatCBFFullPilatus import FormatCBFFullPilatus
from dxtbx.masking import GoniometerMaskerFactory


class FormatCBFFullPilatusDLS6MSN126(FormatCBFFullPilatus):
    """An image reading class for full CBF format images from Pilatus
    detectors."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        header = FormatCBFFullPilatus.get_cbf_header(image_file)

        for record in header.split("\n"):
            if (
                "# Detector" in record
                and "PILATUS" in record
                and "S/N 60-0126" in header
            ):
                return True

        return False

    @staticmethod
    def has_dynamic_shadowing(**kwargs):
        dynamic_shadowing = kwargs.get("dynamic_shadowing", False)
        if dynamic_shadowing in (libtbx.Auto, "Auto"):
            return True
        return dynamic_shadowing

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        self._dynamic_shadowing = self.has_dynamic_shadowing(**kwargs)
        super().__init__(image_file, **kwargs)

    def get_goniometer_shadow_masker(self, goniometer=None):
        if goniometer is None:
            goniometer = self.get_goniometer()

        if not goniometer or not len(self.get_goniometer().get_names()) == 3:
            return None  # No dynamic shadowing available

        if goniometer.get_names()[1] == "GON_CHI":
            # SmarGon
            return GoniometerMaskerFactory.smargon(goniometer)

        else:
            raise RuntimeError(
                "Don't understand this goniometer: %s" % list(goniometer.get_names())
            )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFFullPilatus.understand(arg))



from __future__ import annotations

from os.path import join

import h5py
import numpy as np
from dials.array_family import flex

from dxtbx import flumpy
from dxtbx.format.FormatMANDI import FormatMANDI
from dxtbx.model.beam import BeamFactory, PolychromaticBeam, Probe
from dxtbx.model.scan import Scan, ScanFactory


class FormatMANDILaue(FormatMANDI):
    """
    Class to read MANDI data flattened along the ToF dimension
    to approximate data expected from IMAGINE-X
    https://neutrons.ornl.gov/imagine

    To add Laue data to a MANDI NeXus file:
    >>> from dxtbx.format.FormatMANDILaue import FormatMANDILaue
    >>> nxs_file = "/path/to/file/example.nxs.h5"
    >>> FormatMANDILaue.add_laue_data_to_nxs_file(nxs_file)
    """

    @staticmethod
    def understand(image_file: str) -> bool:
        try:
            return FormatMANDILaue.is_mandi_laue_file(image_file)
        except OSError:
            return False

    @staticmethod
    def is_mandi_laue_file(image_file: str) -> bool:
        with h5py.File(image_file, "r") as handle:
            # File is not empty
            if len(handle) == 0:
                return False
            base_entry = list(handle.keys())[0]
            # Has a name entry
            if f"{base_entry}/instrument/name" not in handle:
                return False
            # Name can be decoded as is MANDI
            try:
                name = handle[f"{base_entry}/instrument/name"][0].decode()
                if name == "MANDI" and "bank10_events" in handle[base_entry].keys():
                    # Laue data present
                    return "laue_data" in handle[f"{base_entry}/bank10_events"]

            except (ValueError, IndexError):
                return False

    def get_raw_data(self, index: int) -> tuple[flex.int]:
        raw_data = []
        for panel_name in self._get_panel_names():
            panel_data = self.nxs_file[self._base_entry][f"{panel_name}_events"][
                "laue_data"
            ][:]
            panel_data = panel_data.astype(np.int32)
            raw_data.append(flumpy.from_numpy(np.ascontiguousarray(panel_data)))
        return tuple(raw_data)

    def get_scan(self, index: int | None = None) -> Scan:
        image_range = (1, self.get_num_images())
        return ScanFactory.make_scan_from_properties(
            image_range=image_range, properties={}
        )

    def get_num_images(self) -> int:
        return 1

    def get_beam(self, index: int | None = None) -> PolychromaticBeam:
        direction = self._get_sample_to_source_direction()
        distance = self._get_sample_to_source_distance()
        wavelength_range = (2.0, 4.0)
        return BeamFactory.make_polychromatic_beam(
            direction=direction,
            sample_to_source_distance=distance,
            probe=Probe.neutron,
            wavelength_range=wavelength_range,
        )

    @staticmethod
    def add_laue_data_to_nxs_file(
        nxs_file_path: str,
        output_name: str = "laue_data",
        panel_size: tuple[int, int] = (256, 256),  # (px)
        compress=True,
    ) -> None:
        """
        Extracts event data and writes out laue data for each panel in
        nxs_file_path
        """

        print(f"Writing laue data in {nxs_file_path}")

        nxs_file = h5py.File(nxs_file_path, "r+")
        base_dir = list(nxs_file.keys())[0]

        panel_names = FormatMANDI.get_panel_names(nxs_file)
        for panel_name in panel_names:
            output_path = join(base_dir, panel_name)
            output_path = join(output_path, output_name)
            panel_data = FormatMANDILaue.generate_laue_data_for_panel(
                nxs_file, panel_size, panel_name
            )
            if compress:
                nxs_file.create_dataset(
                    output_path, data=panel_data, compression="gzip"
                )
            else:
                nxs_file.create_dataset(output_path, data=panel_data)

        nxs_file.close()

    @staticmethod
    def generate_laue_data_for_panel(
        nxs_file: h5py.File,
        panel_size: tuple[int, int],
        panel_name: str,
    ) -> np.array:
        """
        Generates laue data for a given panel
        """

        ## Get panel data
        panel_number = FormatMANDI.get_panel_number(panel_name)
        # Actual pixel ids, starting from bottom left and going up y axis first
        event_id = nxs_file[f"entry/{panel_name}/event_id"][:]

        # event_ids are given with an offset
        event_id_offset = panel_number * panel_size[0] * panel_size[1]
        corrected_event_id = event_id - event_id_offset

        num_pixels = panel_size[0] * panel_size[1]

        counts = np.bincount(corrected_event_id)

        # Pad if no events for pixels outside of max event id
        if len(counts) < num_pixels:
            counts = np.pad(counts, (0, num_pixels - len(counts)), mode="constant")
        else:
            counts = counts[:num_pixels]

        laue_array = np.reshape(counts, panel_size).T

        return laue_array



"""
An implementation of the SMV image reader for ADSC images. Inherits from
FormatSMVADSC, customised for example on ALS beamline 8.3.1 from back in the
day which had its own way of recording beam centre.
"""

from __future__ import annotations

import dateutil.parser
from dateutil.tz import gettz

from dxtbx.format.FormatSMVADSCSN import FormatSMVADSCSN


class FormatSMVADSCSN442(FormatSMVADSCSN):
    """A class for reading SMV format ADSC images, and correctly constructing
    a model for the experiment from this, for instrument number 442."""

    @staticmethod
    def understand(image_file):
        """Check to see if this is ADSC SN 442."""

        # check this is detector serial number 442

        size, header = FormatSMVADSCSN.get_smv_header(image_file)

        return int(header["DETECTOR_SN"]) == 442

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        # this detector lived two lives - it moved on May 14, 2007 to ALS 5.0.1
        # where the denzo beam centre was removed

        # Format will be "%a %b %d %H:%M:%S %Y" or "%a %b %d %H:%M:%S %Z %Y"
        date_str = self._header_dictionary["DATE"]

        # Need to add extra time zones in case of running outside of PDT or PST
        tzinfos = {
            "PDT": gettz("PST8PDT"),
            "PST": gettz("PST8PDT"),
        }
        tm = dateutil.parser.parse(date_str, tzinfos=tzinfos).timetuple()

        als831 = True
        if tm.tm_year > 2006 or (tm.tm_year == 2006 and tm.tm_yday >= 134):
            als831 = False

        if als831:
            beam_x = float(self._header_dictionary["DENZO_X_BEAM"])
            beam_y = float(self._header_dictionary["DENZO_Y_BEAM"])
        else:
            beam_x = float(self._header_dictionary["BEAM_CENTER_X"])
            beam_y = float(self._header_dictionary["BEAM_CENTER_Y"])

        distance = float(self._header_dictionary["DISTANCE"])
        pixel_size = float(self._header_dictionary["PIXEL_SIZE"])
        image_size = (
            float(self._header_dictionary["SIZE1"]),
            float(self._header_dictionary["SIZE2"]),
        )
        trusted_range = self._adsc_trusted_range()

        return self._detector_factory.simple(
            "CCD",
            distance,
            (beam_y, beam_x),
            "+x",
            "-y",
            (pixel_size, pixel_size),
            image_size,
            trusted_range,
            [],
            gain=self._adsc_module_gain(),
            pedestal=40,
        )
